import logging
import os
import uuid
from typing import Dict, List, Optional

import toml
from frozendict import frozendict

from .storage import get_data_manager

# We use these globals for lazy initialization, but pylint doesn't like that.
# pylint: disable=global-statement, invalid-name
_config_paths = _val.split(',') if (_val := os.getenv('CPG_CONFIG_PATH')) else []
_config: Optional[frozendict] = None  # Cached config, initialized lazily.


def _validate_configs(config_paths: List[str]) -> None:
    if [p for p in config_paths if not p.endswith('.toml')]:
        raise ValueError(f'All config files must have ".toml" extensions: {config_paths}')
    if bad_files := [p for p in config_paths if not get_job_config(p)]:
        raise ValueError(f'Some config files do not exist: {bad_files}')


def set_config_paths(config_paths: List[str]) -> None:
    """Sets the config paths that are used by subsequent calls to get_config.

    If this isn't called, the value of the CPG_CONFIG_PATH environment variable is used
    instead.

    Parameters
    ----------
    config_paths: list[str]
        A list of cloudpathlib-compatible paths to TOML files containing configurations.
    """
    global _config_paths, _config
    if _config_paths != config_paths:
        _validate_configs(config_paths)
        _config_paths = config_paths
        os.environ['CPG_CONFIG_PATH'] = ','.join(_config_paths)
        _config = None  # Make sure the config gets reloaded.


def get_config() -> frozendict:
    """Returns the configuration dictionary.

    Call `set_config_paths` beforehand to override the default path.
    See `read_configs` for the path value semantics.

    Notes
    -----
    Caches the result based on the config paths alone.

    Returns
    -------
    dict
    """

    global _config
    if _config is None:  # Lazily initialize the config.
        assert (
            _config_paths
        ), 'Either set the CPG_CONFIG_PATH environment variable or call set_config_paths'

        _config = read_configs(_config_paths)

        # Print the config content, which is helpful for debugging.
        print(
            f'Configuration at {",".join(_config_paths)}:\n{toml.dumps(dict(_config))}'
        )

    return _config


def read_configs(config_paths: List[str]) -> frozendict:
    """Creates a merged configuration from the given config paths.

    For a list of configurations (e.g. ['base.toml', 'override.toml']), the
    configurations get applied from left to right. I.e. the first config gets updated by
    values of the second config, etc.

    Loads `config-template.toml` first as a baseline config.

    Examples
    --------
    Here's a typical configuration file in TOML format:

    [hail]
    billing_project = "tob-wgs"
    bucket = "cpg-tob-wgs-hail"

    [workflow]
    access_level = "test"
    dataset = "tob-wgs"
    dataset_gcp_project = "tob-wgs"
    driver_image = "australia-southeast1-docker.pkg.dev/analysis-runner/images/driver:36c6d4548ef347f14fd34a5b58908057effcde82-hail-ad1fc0e2a30f67855aee84ae9adabc3f3135bd47"
    image_registry_prefix = "australia-southeast1-docker.pkg.dev/cpg-common/images"
    reference_prefix = "gs://cpg-reference"
    output_prefix = "plasma/chr22/v6"

    >>> from cpg_utils.config import get_config
    >>> get_config()['workflow']['dataset']
    'tob-wgs'

    Returns
    -------
    dict
    """

    config: dict = get_base_config()
    for path in config_paths:
        update_dict(config, get_job_config(path))
    return frozendict(config)


def get_base_config() -> Dict:
    """Read in base config toml."""
    template_path = os.path.join(os.path.dirname(__file__), "config-template.toml")
    with open(template_path, "r") as base_config:
        config = toml.loads(base_config.read())
    return config


def get_job_config(config_name: str) -> Optional[Dict]:
    """Reads the given config name from storage to a dictionary."""
    # TODO GRS need deployment-specific location for GCP.
    config_bytes = get_data_manager().get_blob(None, "config", config_name)
    return toml.loads(config_bytes.decode("utf-8")) if config_bytes else None


def set_job_config(config: Dict) -> str:
    """Writes the given config dictionary to a blob and returns its unique name."""
    # TODO GRS need deployment-specific location for GCP.
    config_name = str(uuid.uuid4()) + ".toml"
    contents = toml.dumps(config).encode("utf-8")
    get_data_manager().set_blob(None, "config", config_name, contents)
    return config_name


def remote_tmpdir(hail_bucket: Optional[str] = None) -> str:
    """Returns the remote_tmpdir to use for Hail initialization for a particular dataset.

    If `hail_bucket` is not specified explicitly, requires the `hail/bucket` config variable to be set.
    """
    bucket = hail_bucket or get_config().get("hail", {}).get("bucket")
    assert bucket, f"hail_bucket was not set by argument or configuration"
    return f"{bucket}/batch-tmp"


def update_dict(d1: Dict, d2: Dict) -> None:
    """Updates the d1 dict with the values from the d2 dict recursively in-place."""
    for k, v2 in d2.items():
        v1 = d1.get(k)
        if isinstance(v1, dict) and isinstance(v2, dict):
            update_dict(v1, v2)
        else:
            d1[k] = v2


_validate_configs(_config_paths)



"""
Helpers to communicate with the sample-metadata database.
"""

import logging
import pprint
import traceback
from dataclasses import dataclass
from enum import Enum
from typing import Optional

from cpg_utils.config import get_config
from cpg_utils import Path, to_path

from sample_metadata import models
from sample_metadata.apis import (
    SampleApi,
    SequenceApi,
    AnalysisApi,
    ParticipantApi,
    FamilyApi,
)
from sample_metadata.exceptions import ApiException

from cpg_utils.workflows.utils import exists
from cpg_utils.workflows.filetypes import (
    FastqPair,
    CramPath,
    BamPath,
    AlignmentInput,
    FastqPairs,
)


_metamist: Optional['Metamist'] = None


def get_metamist() -> 'Metamist':
    """Return the cohort object"""
    global _metamist
    if not _metamist:
        _metamist = Metamist()
    return _metamist


class MetamistError(Exception):
    """
    Error while interacting with Metamist.
    """

    pass


class AnalysisStatus(Enum):
    """
    Corresponds to metamist Analysis statuses:
    https://github.com/populationgenomics/sample-metadata/blob/dev/models/enums/analysis.py#L14-L21
    """

    QUEUED = 'queued'
    IN_PROGRESS = 'in-progress'
    FAILED = 'failed'
    COMPLETED = 'completed'
    UNKNOWN = 'unknown'

    @staticmethod
    def parse(name: str) -> 'AnalysisStatus':
        """
        Parse str and create a AnalysisStatus object
        """
        return {v.value: v for v in AnalysisStatus}[name.lower()]


class AnalysisType(Enum):
    """
    Corresponds to metamist Analysis types:
    https://github.com/populationgenomics/sample-metadata/blob/dev/models/enums
    /analysis.py#L4-L11

    Re-defined in a separate module to decouple from the main sample-metadata module,
    so decorators can use `@stage(analysis_type=AnalysisType.QC)` without importing
    the sample-metadata package.
    """

    QC = 'qc'
    JOINT_CALLING = 'joint-calling'
    GVCF = 'gvcf'
    CRAM = 'cram'
    CUSTOM = 'custom'
    ES_INDEX = 'es-index'

    @staticmethod
    def parse(val: str) -> 'AnalysisType':
        """
        Parse str and create a AnalysisStatus object
        """
        d = {v.value: v for v in AnalysisType}
        if val not in d:
            raise MetamistError(
                f'Unrecognised analysis type {val}. Available: {list(d.keys())}'
            )
        return d[val.lower()]


@dataclass
class Analysis:
    """
    Sample metadata DB Analysis entry.

    See the sample-metadata package for more details:
    https://github.com/populationgenomics/sample-metadata
    """

    id: int
    type: AnalysisType
    status: AnalysisStatus
    sample_ids: set[str]
    output: Path | None
    meta: dict

    @staticmethod
    def parse(data: dict) -> 'Analysis':
        """
        Parse data to create an Analysis object.
        """
        req_keys = ['id', 'type', 'status']
        if any(k not in data for k in req_keys):
            for key in req_keys:
                if key not in data:
                    logging.error(f'"Analysis" data does not have {key}: {data}')
            raise ValueError(f'Cannot parse metamist Sequence {data}')

        output = data.get('output')
        if output:
            output = to_path(output)

        a = Analysis(
            id=int(data['id']),
            type=AnalysisType.parse(data['type']),
            status=AnalysisStatus.parse(data['status']),
            sample_ids=set(data.get('sample_ids', [])),
            output=output,
            meta=data.get('meta') or {},
        )
        return a


class Metamist:
    """
    Communication with metamist.
    """

    def __init__(self):
        self.default_dataset: str = get_config()['workflow']['dataset']
        self.sapi = SampleApi()
        self.aapi = AnalysisApi()
        self.seqapi = SequenceApi()
        self.papi = ParticipantApi()
        self.fapi = FamilyApi()

    def update_analysis(self, analysis: Analysis, status: AnalysisStatus):
        """
        Update "status" of an Analysis entry.
        """
        try:
            self.aapi.update_analysis_status(
                analysis.id,
                models.AnalysisUpdateModel(status=models.AnalysisStatus(status.value)),
            )
        except ApiException:
            traceback.print_exc()
        analysis.status = status

    def find_joint_calling_analysis(
        self,
        sample_ids: list[str],
        dataset: str | None = None,
    ) -> Analysis | None:
        """
        Query the DB to find the last completed joint-calling analysis for the samples.
        """
        project = dataset or self.default_dataset
        if get_config()['workflow']['access_level'] == 'test':
            project += '-test'
        try:
            data = self.aapi.get_latest_complete_analysis_for_type(
                project=project,
                analysis_type=models.AnalysisType('joint-calling'),
            )
        except ApiException:
            return None
        a = Analysis.parse(data)
        if not a:
            return None
        assert a.type == AnalysisType.JOINT_CALLING, data
        assert a.status == AnalysisStatus.COMPLETED, data
        if a.sample_ids != set(sample_ids):
            return None
        return a

    def find_analyses_by_sid(
        self,
        sample_ids: list[str],
        analysis_type: AnalysisType,
        analysis_status: AnalysisStatus = AnalysisStatus.COMPLETED,
        meta: dict | None = None,
        dataset: str | None = None,
    ) -> dict[str, Analysis]:
        """
        Query the DB to find the last completed analysis for the type and samples,
        one Analysis object per sample. Assumes the analysis is defined for a single
        sample (e.g. cram, gvcf).
        """
        dataset = dataset or self.default_dataset
        project = dataset or self.default_dataset
        if get_config()['workflow']['access_level'] == 'test':
            project += '-test'

        analysis_per_sid: dict[str, Analysis] = dict()

        logging.info(f'Querying {analysis_type} analysis entries for {project}...')
        datas = self.aapi.query_analyses(
            models.AnalysisQueryModel(
                projects=[project],
                sample_ids=sample_ids,
                type=models.AnalysisType(analysis_type.value),
                status=models.AnalysisStatus(analysis_status.value),
                meta=meta or {},
            )
        )

        for data in datas:
            a = Analysis.parse(data)
            if not a:
                continue
            assert a.status == AnalysisStatus.COMPLETED, data
            assert a.type == analysis_type, data
            assert len(a.sample_ids) == 1, data
            analysis_per_sid[list(a.sample_ids)[0]] = a
        logging.info(
            f'Querying {analysis_type} analysis entries for {project}: '
            f'found {len(analysis_per_sid)}'
        )
        return analysis_per_sid

    def create_analysis(
        self,
        output: Path | str,
        type_: str | AnalysisType,
        status: str | AnalysisStatus,
        sample_ids: list[str],
        dataset: str | None = None,
        meta: dict | None = None,
    ) -> int | None:
        """
        Tries to create an Analysis entry, returns its id if successful.
        """
        dataset = dataset or self.default_dataset
        project = dataset or self.default_dataset
        if get_config()['workflow']['access_level'] == 'test':
            project += '-test'

        if isinstance(type_, AnalysisType):
            type_ = type_.value
        if isinstance(status, AnalysisStatus):
            status = status.value

        am = models.AnalysisModel(
            type=models.AnalysisType(type_),
            status=models.AnalysisStatus(status),
            output=str(output),
            sample_ids=list(sample_ids),
            meta=meta or {},
        )
        try:
            aid = self.aapi.create_new_analysis(project=project, analysis_model=am)
        except ApiException:
            traceback.print_exc()
            return None
        else:
            logging.info(
                f'Created Analysis(id={aid}, type={type_}, status={status}, '
                f'output={str(output)}) in {project}'
            )
            return aid

    def process_existing_analysis(
        self,
        sample_ids: list[str],
        completed_analysis: Analysis | None,
        analysis_type: str,
        expected_output_fpath: Path,
        dataset: str | None = None,
    ) -> Path | None:
        """
        Checks whether existing analysis exists, and output matches the expected output
        file. Invalidates bad analysis by setting status=failure, and submits a
        status=completed analysis if the expected output already exists.

        Returns the path to the output if it can be reused, otherwise None.

        @param sample_ids: sample IDs to pull the analysis for
        @param completed_analysis: existing completed analysis of this type for these
        samples
        @param analysis_type: cram, gvcf, joint_calling
        @param expected_output_fpath: where the workflow expects the analysis output
        file to sit on the bucket (will invalidate the analysis when it doesn't match)
        @param dataset: the name of the dataset to create a new analysis
        @return: path to the output if it can be reused, otherwise None
        """
        label = f'type={analysis_type}'
        if len(sample_ids) > 1:
            label += f' for {", ".join(sample_ids)}'

        found_output_fpath: Path | None = None
        if not completed_analysis:
            logging.warning(
                f'Not found completed analysis {label} for '
                f'{f"sample {sample_ids}" if len(sample_ids) == 1 else f"{len(sample_ids)} samples" }'
            )
        elif not completed_analysis.output:
            logging.error(
                f'Found a completed analysis {label}, '
                f'but the "output" field does not exist or empty'
            )
        else:
            found_output_fpath = completed_analysis.output
            if found_output_fpath != expected_output_fpath:
                logging.error(
                    f'Found a completed analysis {label}, but the "output" path '
                    f'{found_output_fpath} does not match the expected path '
                    f'{expected_output_fpath}'
                )
                found_output_fpath = None
            elif not exists(found_output_fpath):
                logging.error(
                    f'Found a completed analysis {label}, '
                    f'but the "output" file {found_output_fpath} does not exist'
                )
                found_output_fpath = None

        # completed and good exists, can reuse
        if found_output_fpath:
            logging.info(
                f'Completed analysis {label} exists, '
                f'reusing the result {found_output_fpath}'
            )
            return found_output_fpath

        # can't reuse, need to invalidate
        if completed_analysis:
            logging.warning(
                f'Invalidating the analysis {label} by setting the status to "failure", '
                f'and resubmitting the analysis.'
            )
            self.update_analysis(completed_analysis, status=AnalysisStatus.FAILED)

        # can reuse, need to create a completed one?
        if exists(expected_output_fpath):
            logging.info(
                f'Output file {expected_output_fpath} already exists, so creating '
                f'an analysis {label} with status=completed'
            )
            self.create_analysis(
                type_=analysis_type,
                output=expected_output_fpath,
                status='completed',
                sample_ids=sample_ids,
                dataset=dataset or self.default_dataset,
            )
            return expected_output_fpath

        # proceeding with the standard workflow (creating status=queued, submitting jobs)
        else:
            logging.info(
                f'Expected output file {expected_output_fpath} does not exist, '
                f'so queueing analysis {label}'
            )
            return None

    def get_ped_entries(self, dataset: str | None = None) -> list[dict[str, str]]:
        """
        Retrieve PED lines for a specified SM project, with external participant IDs.
        """
        dataset = dataset or self.default_dataset

        families = self.fapi.get_families(dataset)
        family_ids = [family['id'] for family in families]
        ped_entries = self.fapi.get_pedigree(
            internal_family_ids=family_ids,
            response_type='json',
            project=dataset,
            replace_with_participant_external_ids=True,
        )

        return ped_entries


@dataclass
class Sequence:
    """
    Metamist "Sequence" entry.

    See metamist for more details:
    https://github.com/populationgenomics/sample-metadata
    """

    id: str
    sample_id: str
    meta: dict
    sequencing_type: str
    alignment_input: AlignmentInput | None = None

    @staticmethod
    def parse(data: dict, check_existence: bool = False) -> 'Sequence':
        """
        Create from a dictionary.
        """
        req_keys = ['id', 'sample_id', 'meta']
        if any(k not in data for k in req_keys):
            for key in req_keys:
                if key not in data:
                    logging.error(f'"Sequence" data does not have {key}: {data}')
            raise ValueError(f'Cannot parse metamist Sequence {data}')

        sample_id = data['sample_id']
        sequencing_type = data['type']
        assert sequencing_type, data
        mm_seq = Sequence(
            id=data['id'],
            sample_id=sample_id,
            meta=data['meta'],
            sequencing_type=sequencing_type,
        )
        if data['meta'].get('reads'):
            if alignment_input := Sequence._parse_reads(
                sample_id=sample_id,
                meta=data['meta'],
                check_existence=check_existence,
            ):
                mm_seq.alignment_input = alignment_input
        else:
            logging.warning(
                f'{sample_id} sequence: no meta/reads found with FASTQ information'
            )
        return mm_seq

    @staticmethod
    def _parse_reads(  # pylint: disable=too-many-return-statements
        sample_id: str,
        meta: dict,
        check_existence: bool,
    ) -> AlignmentInput | None:
        """
        Parse a AlignmentInput object from the meta dictionary.

        @param check_existence: check if fastq/crams exist on buckets.
        Default value is pulled from self.metamist and can be overridden.
        """
        reads_data = meta.get('reads')
        reads_type = meta.get('reads_type')
        reference_assembly = meta.get('reference_assembly', {}).get('location')

        if not reads_data:
            logging.error(f'{sample_id}: no "meta/reads" field in meta')
            return None
        if not reads_type:
            logging.error(f'{sample_id}: no "meta/reads_type" field in meta')
            return None
        supported_types = ('fastq', 'bam', 'cram')
        if reads_type not in supported_types:
            logging.error(
                f'{sample_id}: ERROR: "reads_type" is expected to be one of '
                f'{supported_types}'
            )
            return None

        if reads_type in ('bam', 'cram'):
            if len(reads_data) > 1:
                logging.error(f'{sample_id}: supporting only single bam/cram input')
                return None

            location = reads_data[0]['location']
            if not (location.endswith('.cram') or location.endswith('.bam')):
                logging.error(
                    f'{sample_id}: ERROR: expected the file to have an extension '
                    f'.cram or .bam, got: {location}'
                )
                return None
            if get_config()['workflow']['access_level'] == 'test':
                location = location.replace('-main-upload/', '-test-upload/')
            if check_existence and not exists(location):
                logging.error(
                    f'{sample_id}: ERROR: index file does not exist: {location}'
                )
                return None

            # Index:
            index_location = None
            if reads_data[0].get('secondaryFiles'):
                index_location = reads_data[0]['secondaryFiles'][0]['location']
                if (
                    location.endswith('.cram')
                    and not index_location.endswith('.crai')
                    or location.endswith('.bai')
                    and not index_location.endswith('.bai')
                ):
                    logging.error(
                        f'{sample_id}: ERROR: expected the index file to have an extension '
                        f'.crai or .bai, got: {index_location}'
                    )
                if get_config()['workflow']['access_level'] == 'test':
                    index_location = index_location.replace(
                        '-main-upload/', '-test-upload/'
                    )
                if check_existence and not exists(index_location):
                    logging.error(
                        f'{sample_id}: ERROR: index file does not exist: {index_location}'
                    )
                    return None

            if location.endswith('.cram'):
                return CramPath(
                    location,
                    index_path=index_location,
                    reference_assembly=reference_assembly,
                )
            else:
                assert location.endswith('.bam')
                return BamPath(location, index_path=index_location)

        else:
            fastq_pairs = FastqPairs()
            for lane_pair in reads_data:
                if len(lane_pair) != 2:
                    raise ValueError(
                        f'Sequence data for sample {sample_id} is incorrectly '
                        f'formatted. Expecting 2 entries per lane (R1 and R2 fastqs), '
                        f'but got {len(lane_pair)}. '
                        f'Read data: {pprint.pformat(reads_data)}'
                    )
                if get_config()['workflow']['access_level'] == 'test':
                    lane_pair[0]['location'] = lane_pair[0]['location'].replace(
                        '-main-upload/', '-test-upload/'
                    )
                    lane_pair[1]['location'] = lane_pair[1]['location'].replace(
                        '-main-upload/', '-test-upload/'
                    )
                if check_existence and not exists(lane_pair[0]['location']):
                    logging.error(
                        f'{sample_id}: ERROR: read 1 file does not exist: '
                        f'{lane_pair[0]["location"]}'
                    )
                    return None
                if check_existence and not exists(lane_pair[1]['location']):
                    logging.error(
                        f'{sample_id}: ERROR: read 2 file does not exist: '
                        f'{lane_pair[1]["location"]}'
                    )
                    return None

                fastq_pairs.append(
                    FastqPair(
                        to_path(lane_pair[0]['location']),
                        to_path(lane_pair[1]['location']),
                    )
                )

            return fastq_pairs



import azure.core.exceptions
import azure.identity
import azure.storage.blob
import google.cloud.storage
import pytest
from cpg_utils.deploy_config import set_deploy_config_from_env
from cpg_utils.storage import clear_data_manager, DataManager, get_data_manager, get_dataset_bucket_url
from cpg_utils.job_config import get_config, remote_tmpdir, set_config_paths, set_job_config, _validate_configs


class MockStorageResponse:
    def __init__(self, blob_contents):
        self.value = bytes(blob_contents, "UTF-8")
    def download_as_bytes(self):
        return self.value
    def readall(self):
        return self.value
    def open(self, mode):
        assert mode == "wb"
        return self
    def __enter__(self):
        return self
    def write(self, contents):
        pass
    def __exit__(self, type, value, traceback):
        pass

class MockStorageClientGCP:
    def bucket(self, bucket_name):
        assert bucket_name in ["cpg-dataset0-main-read", "cpg-config"]
        return self
    def get_blob(self, blob_path):
        if blob_path in ["missing.json", "missing.toml"]:
            return None
        if blob_path == "config.toml":
            return MockStorageResponse(
                """[hail]
                billing_project = "dataset0"
                bucket = "gs://cpg-dataset0-hail"
                [workflow]
                access_level = "pytestgcp"
                """
            )
        assert blob_path == "exists.json"
        return MockStorageResponse("GCP BLOB CONTENTS")
    def blob(self, blob_path):
        return MockStorageResponse("nothing")

class MockStorageClientAzure:
    def __init__(self, *args, **kwargs):
        if len(args) > 2:
            assert args[0] in ["https://dataset1_idsa.blob.core.windows.net", "https://analysis-runnersa.blob.core.windows.net"]
            assert args[1] in ["cpg-dataset1-main-read", "cpg-config"]
            self.path = args[2]
    def download_blob(self):
        if self.path == "missing.json":
            raise azure.core.exceptions.ResourceNotFoundError()
        if self.path == "config.toml":
            return MockStorageResponse(
                """[hail]
                billing_project = "dataset1"
                bucket = "hail-az://dataset1_idsa/cpg-dataset1-hail"
                [workflow]
                access_level = "pytestaz"
                """
            )
        assert self.path == "exists.json"
        return MockStorageResponse("AZURE BLOB CONTENTS")
    def upload_blob(self, data: bytes, overwrite: bool):
        pass

@pytest.fixture
def mock_config_fixture(json_load):
    def mock_get_server_config():
        return json_load("server_config_01.json")
    return mock_get_server_config


def test_gcp_storage(monkeypatch):
    monkeypatch.setattr(google.cloud.storage, "Client", MockStorageClientGCP)
    monkeypatch.setenv("CLOUD", "gcp")
    set_deploy_config_from_env()
    sm = get_data_manager()

    assert sm.get_blob("dataset0", "main-read", "exists.json").decode("UTF-8") == "GCP BLOB CONTENTS"
    assert sm.get_blob("dataset0", "main-read", "missing.json") == None
    sm.set_blob("dataset0", "main-read", "exists.json", bytes("GCP BLOB CONTENTS", "UTF-8"))

    with pytest.raises(ValueError) as e:
        _validate_configs(["test.txt"])
    with pytest.raises(ValueError) as e:
        _validate_configs(["missing.toml"])

    set_config_paths(["config.toml"])
    assert get_config()["workflow"]["access_level"] == "pytestgcp"
    set_job_config({"workflow":{"access_level":"pytestgcp"}})

    assert get_dataset_bucket_url("dataset0", "test") == "gs://cpg-dataset0-test"
    assert remote_tmpdir() == "gs://cpg-dataset0-hail/batch-tmp"


def test_azure_storage(monkeypatch, mock_config_fixture):
    monkeypatch.setattr("cpg_utils.storage.get_server_config", mock_config_fixture)
    monkeypatch.setattr(azure.identity, "DefaultAzureCredential", MockStorageClientAzure)
    monkeypatch.setattr(azure.storage.blob, "BlobClient", MockStorageClientAzure)
    monkeypatch.setenv("CLOUD", "azure")
    monkeypatch.delenv("CPG_DEPLOY_CONFIG", raising=False)
    set_deploy_config_from_env()
    sm = DataManager.get_data_manager()

    assert sm.get_blob("dataset1", "main-read", "exists.json").decode("UTF-8") == "AZURE BLOB CONTENTS"
    assert sm.get_blob("dataset1", "main-read", "missing.json") == None
    with pytest.raises(ValueError) as e:
        sm.get_blob("dataset0", "main-read", "exists.json")
        assert "No such dataset in server config" in str(e.value)
    sm.set_blob("dataset1", "main-read", "exists.json", bytes("AZURE BLOB CONTENTS", "UTF-8"))
    with pytest.raises(ValueError) as e:
        sm.set_blob("dataset0", "main-read", "exists.json", bytes("AZURE BLOB CONTENTS", "UTF-8"))
        assert "No such dataset in server config" in str(e.value)

    clear_data_manager()
    set_config_paths([])
    set_config_paths(["config.toml"])
    assert get_config()["workflow"]["access_level"] == "pytestaz"

    assert sm.get_dataset_bucket_url("dataset1", "test") == "hail-az://dataset1_idsa/cpg-dataset1-test"
    assert remote_tmpdir("hail-az://dataset1_idsa/cpg-dataset1-hail") == \
        "hail-az://dataset1_idsa/cpg-dataset1-hail/batch-tmp"



"""
Test Hail Query functions.
"""

import hail as hl
import toml

from cpg_utils import to_path
from cpg_utils.config import set_config_paths
from cpg_utils.workflows.batch import get_batch
from cpg_utils.workflows.inputs import get_cohort
from cpg_utils.workflows.targets import Sample, Cohort
from cpg_utils.workflows.utils import timestamp
from cpg_utils.workflows.workflow import (
    Workflow,
    SampleStage,
    StageInput,
    StageOutput,
    ExpectedResultT,
    CohortStage,
    stage,
)
from cpg_utils.hail_batch import dataset_path, command

tmp_dir_path = to_path(__file__).parent / 'results' / timestamp()
tmp_dir_path = tmp_dir_path.absolute()
tmp_dir_path.mkdir(parents=True, exist_ok=True)

config_toml = f"""
[workflow]
dataset_gcp_project = 'fewgenomes'
access_level = 'test'
dataset = 'fewgenomes'
check_inputs = false
check_intermediates = false
check_expected_outputs = false
path_scheme = 'local'
local_dir = '{str(tmp_dir_path)}'

[hail]
billing_project = 'fewgenomes'
delete_scratch_on_exit = false
backend = 'local'
"""


def _set_config():
    config_path = tmp_dir_path / 'config.toml'
    with config_path.open('w') as f:
        toml.dump(toml.loads(config_toml), f)
    set_config_paths([str(config_path)])


_set_config()


def test_batch_job():
    """
    Test creating a job and running a batch.
    """
    b = get_batch('Test batch job')
    j1 = b.new_job('Jo b1')
    text = 'success'
    cmd = f"""\
    echo {text} > {j1.output}
    """
    j1.command(command(cmd))
    output1_path = dataset_path('output1.txt')
    b.write_output(j1.output, str(output1_path))

    j2 = b.new_job('Job 2')
    j2.command(f'touch {j2.output}')
    j2.command(f'cat {b.read_input(output1_path)} >> {j2.output}')
    j2.depends_on(j1)
    output2_path = dataset_path('output2.txt')
    b.write_output(j2.output, str(output2_path))

    b.run()
    with to_path(output2_path).open() as fh:
        assert fh.read().strip() == text


def test_batch_python_job():
    """
    Testing calling a python job.
    """
    b = get_batch('Test batch python job')
    j = b.new_python_job('Test python job')

    input_tsv_path = to_path(dataset_path('input.tsv'))
    input_tsv_path.parent.mkdir(parents=True, exist_ok=True)
    with input_tsv_path.open('w') as f:
        f.write('col1\tcol2\n1\t2')

    def query_fn(tsv_path: str, out_ht_path: str):
        ht = hl.import_table(tsv_path, types={'col1': hl.tint, 'col2': hl.tint})
        ht.show()
        ht = ht.annotate(col3=ht.col1 + ht.col2)
        ht.write(out_ht_path, overwrite=True)

    output_ht_path = dataset_path('output.ht')
    j.call(query_fn, str(input_tsv_path), output_ht_path)
    b.run()

    hl.init_local(log=dataset_path('hail-log.txt'))
    result = hl.read_table(str(output_ht_path)).col3.collect()[0]
    assert result == 3, result


def test_cohort(mocker):
    """
    Testing creating a Cohort object from metamist mocks.
    """

    def mock_get_samples(  # pylint: disable=unused-argument
        *args, **kwargs
    ) -> list[dict]:
        return [
            {'id': 'CPG01', 'external_id': 'SAMPLE1'},
            {'id': 'CPG02', 'external_id': 'SAMPLE2'},
        ]

    def mock_get_sequences_by_sample_ids(  # pylint: disable=unused-argument
        *args, **kwargs
    ) -> list[dict]:
        return [
            {
                'id': 0,
                'sample_id': 'CPG01',
                'type': 'genome',
                'status': 'completed',
                'meta': {'reads': {'location': 'mock'}, 'read_type': 'bam'},
            },
            {
                'id': 1,
                'sample_id': 'CPG02',
                'type': 'genome',
                'status': 'completed',
                'meta': {'reads': {'location': 'mock'}, 'read_type': 'bam'},
            },
        ]

    def mock_get_external_participant_id_to_internal_sample_id(  # pylint: disable=unused-argument
        *args, **kwargs
    ) -> list[list]:
        return [['CPG01', 'PART1'], ['CPG02', 'PART2']]

    def mock_get_families(*args, **kwargs):  # pylint: disable=unused-argument
        return []

    def mock_get_pedigree(*args, **kwargs):  # pylint: disable=unused-argument
        return []

    def mock_query_analyses(*args, **kwargs):  # pylint: disable=unused-argument
        return []

    mocker.patch(
        'sample_metadata.apis.SampleApi.get_samples',
        mock_get_samples,
    )
    mocker.patch(
        'sample_metadata.apis.SequenceApi.get_sequences_by_sample_ids',
        mock_get_sequences_by_sample_ids,
    )
    mocker.patch(
        'sample_metadata.apis.ParticipantApi.get_external_participant_id_to_internal_sample_id',
        mock_get_external_participant_id_to_internal_sample_id,
    )
    mocker.patch(
        'sample_metadata.apis.FamilyApi.get_families',
        mock_get_families,
    )
    mocker.patch(
        'sample_metadata.apis.FamilyApi.get_pedigree',
        mock_get_pedigree,
    )
    mocker.patch(
        'sample_metadata.apis.AnalysisApi.query_analyses',
        mock_query_analyses,
    )

    cohort = get_cohort()
    assert cohort.get_samples()[0].id == 'CPG01'


def test_workflow(mocker):
    """
    Testing running a workflow from a mock cohort.
    """

    def mock_create_cohort() -> Cohort:
        c = Cohort()
        ds = c.create_dataset('my_dataset')
        ds.add_sample('CPG01', external_id='SAMPLE1')
        ds.add_sample('CPG02', external_id='SAMPLE2')
        return c

    mocker.patch('cpg_utils.workflows.inputs.create_cohort', mock_create_cohort)

    output_path = to_path(dataset_path('cohort.tsv'))

    assert len(get_cohort().get_samples()) == 2

    @stage
    class MySampleStage(SampleStage):
        """
        Just a sample-level stage.
        """

        def expected_outputs(self, sample: Sample) -> ExpectedResultT:
            return dataset_path(f'{sample.id}.tsv')

        def queue_jobs(self, sample: Sample, inputs: StageInput) -> StageOutput | None:
            j = self.b.new_job('Sample job', self.get_job_attrs(sample))
            j.command(f'echo {sample.id}_done >> {j.output}')
            self.b.write_output(j.output, str(self.expected_outputs(sample)))
            print(f'Writing to {self.expected_outputs(sample)}')
            return self.make_outputs(sample, self.expected_outputs(sample))

    @stage(required_stages=MySampleStage)
    class MyCohortStage(CohortStage):
        """
        Just a cohort-level stage.
        """

        def expected_outputs(self, cohort: Cohort) -> ExpectedResultT:
            return output_path

        def queue_jobs(self, cohort: Cohort, inputs: StageInput) -> StageOutput | None:
            path_by_sample = inputs.as_path_by_target(MySampleStage)
            assert len(path_by_sample) == len(cohort.get_samples())
            j = self.b.new_job('Cohort job', self.get_job_attrs(cohort))
            j.command(f'touch {j.output}')
            for _, sample_result_path in path_by_sample.items():
                input_file = self.b.read_input(str(sample_result_path))
                j.command(f'cat {input_file} >> {j.output}')
            self.b.write_output(j.output, str(self.expected_outputs(cohort)))
            print(f'Writing to {self.expected_outputs(cohort)}')
            return self.make_outputs(cohort, self.expected_outputs(cohort))

    workflow = Workflow(stages=[MyCohortStage])
    workflow.run()

    print(f'Checking result in {output_path}:')
    with output_path.open() as f:
        result = f.read()
        print(result)
        assert result.split() == ['CPG01_done', 'CPG02_done'], result



import json

import azure.identity
import azure.keyvault.secrets as secrets
import pytest
from cpg_utils.auth import check_dataset_access, check_global_access
from cpg_utils.deploy_config import get_deploy_config, get_server_config, set_deploy_config_from_env
from cpg_utils.secrets import SecretManager
from google.cloud import secretmanager

TEST_SERVER_CONFIG = json.dumps({
	"dataset1": {
		"projectId": "dataset1_id",
		"allowedRepos": ["sample-metadata", "fewgenomes"],
		"testToken": "Hail test SA account",
		"standardToken": "Hail standard SA account",
		"fullToken": "Hail full SA account" 
	}
})


class MockSecretResponse:
    class Payload:
        def __init__(self, secret_value):
            self.data = bytes(secret_value, "UTF-8")
    def __init__(self, secret_value):
        self.payload = MockSecretResponse.Payload(secret_value)
        self.value = secret_value


class MockSecretClientAzure:
    def __init__(self, *args, **kwargs):
        pass
    def get_secret(self, secret_name):
        if secret_name == "server-config":
            return MockSecretResponse(TEST_SERVER_CONFIG)
        if secret_name == "test_name":
            return MockSecretResponse("supersecret in azure")
        if secret_name == "project-creator-users":
            return MockSecretResponse("you@example.com,admin2@test.com")
        assert secret_name == "dataset1-read-members-cache"
        return MockSecretResponse("me@example.com,test2@test.com")


class MockSecretClientGCP:
    def secret_path(self, secret_host, secret_name):
        return secret_host + "/" + secret_name
    def access_secret_version(self, request):
        if request["name"] == "analysis-runner/server-config/versions/latest":
            return MockSecretResponse(TEST_SERVER_CONFIG)
        if request["name"] == "test_host/test_name/versions/latest":
            return MockSecretResponse("supersecret in gcp")
        if request["name"] == "analysis-runner/project-creator-users/versions/latest":
            return MockSecretResponse("you@example.com,admin1@test.com")
        assert request["name"] == "dataset1_id/dataset1-read-members-cache/versions/latest"
        return MockSecretResponse("me@example.com,test1@test.com")


def test_gcp_secret(monkeypatch):
    monkeypatch.delenv("CPG_DEPLOY_CONFIG", raising=False)
    monkeypatch.setattr(secretmanager, "SecretManagerServiceClient", MockSecretClientGCP)
    sm = SecretManager.get_secret_manager("gcp")
    assert sm.read_secret("test_host", "test_name") == "supersecret in gcp"

    monkeypatch.setenv("CLOUD", "gcp")
    set_deploy_config_from_env()
    assert check_dataset_access("dataset1", "test1@test.com", "read") == True
    assert check_dataset_access("dataset1", "test2@test.com", "read") == False
    assert check_dataset_access("dataset2", "test2@test.com", "read") == False
    assert check_global_access("admin1@test.com", "project-creator-users") == True
    assert check_global_access("admin2@test.com", "project-creator-users") == False


def test_azure_secret(monkeypatch):
    monkeypatch.setattr(azure.identity, "DefaultAzureCredential", MockSecretClientAzure)
    monkeypatch.setattr(secrets, "SecretClient", MockSecretClientAzure)
    sm = SecretManager.get_secret_manager("azure")
    assert sm.read_secret("test_host", "test_name") == "supersecret in azure"

    monkeypatch.setenv("CLOUD", "azure")
    set_deploy_config_from_env()
    assert check_dataset_access("dataset1", "test1@test.com", "read") == False
    assert check_dataset_access("dataset1", "test2@test.com", "read") == True
    assert check_dataset_access("dataset2", "test2@test.com", "read") == False
    assert check_global_access("admin1@test.com", "project-creator-users") == False
    assert check_global_access("admin2@test.com", "project-creator-users") == True


def test_server_config(monkeypatch):
    monkeypatch.delenv("CPG_DEPLOY_CONFIG", raising=False)
    monkeypatch.setattr(secretmanager, "SecretManagerServiceClient", MockSecretClientGCP)
    monkeypatch.setenv("CLOUD", "gcp")
    set_deploy_config_from_env()

    scfg1 = get_deploy_config().server_config
    scfg2 = get_server_config()
    assert scfg1 == scfg2



"""Convenience functions related to Hail."""

import asyncio
import inspect
import os
import tempfile
import textwrap
from enum import Enum
from typing import Optional, List, Union
from abc import ABC, abstractmethod

import hail as hl
import hailtop.batch as hb
from hail.utils.java import Env

from cpg_utils.config import get_config
from cpg_utils import to_path, Path


# template commands strings
GCLOUD_AUTH_COMMAND = """\
export GOOGLE_APPLICATION_CREDENTIALS=/gsa-key/key.json
gcloud -q auth activate-service-account \
--key-file=$GOOGLE_APPLICATION_CREDENTIALS
"""


def init_batch(**kwargs):
    """
    Initializes the Hail Query Service from within Hail Batch.
    Requires the `hail/billing_project` and `hail/bucket` config variables to be set.

    Parameters
    ----------
    kwargs : keyword arguments
        Forwarded directly to `hl.init_batch`.
    """
    # noinspection PyProtectedMember
    if Env._hc:  # pylint: disable=W0212
        return  # already initialised
    dataset = get_config()['workflow']['dataset']
    kwargs.setdefault('token', os.environ.get('HAIL_TOKEN'))
    kwargs.setdefault('default_reference', genome_build())
    asyncio.get_event_loop().run_until_complete(
        hl.init_batch(
            billing_project=get_config()['hail']['billing_project'],
            remote_tmpdir=remote_tmpdir(f'cpg-{dataset}-hail'),
            **kwargs,
        )
    )


def copy_common_env(job: hb.batch.job.Job) -> None:
    """Copies common environment variables that we use to run Hail jobs.

    These variables are typically set up in the analysis-runner driver, but need to be
    passed through for "batch-in-batch" use cases.

    The environment variable values are extracted from the current process and
    copied to the environment dictionary of the given Hail Batch job.
    """
    # If possible, please don't add new environment variables here, but instead add
    # config variables.
    for key in ('CPG_CONFIG_PATH',):
        val = os.getenv(key)
        if val:
            job.env(key, val)


def remote_tmpdir(hail_bucket: Optional[str] = None) -> str:
    """Returns the remote_tmpdir to use for Hail initialization.

    If `hail_bucket` is not specified explicitly, requires the `hail/bucket` config variable to be set.
    """
    bucket = hail_bucket or get_config().get('hail', {}).get('bucket')
    assert bucket, f'hail_bucket was not set by argument or configuration'
    return f'gs://{bucket}/batch-tmp'


class PathScheme(ABC):
    """
    Cloud storage path scheme. Constructs full paths to buckets and files.
    """

    @abstractmethod
    def path_prefix(self, dataset: str, category: str) -> str:
        """Build path prefix used in dataset_path"""

    @abstractmethod
    def full_path(self, prefix: str, suffix: str) -> str:
        """Build full path from prefix and suffix"""

    @staticmethod
    def parse(val: str) -> 'PathScheme':
        """Parse subclass name from string"""
        if val == 'gs':
            return GSPathScheme()
        if val == 'hail-az':
            return AzurePathScheme()
        if val == 'local':
            return LocalPathScheme()
        raise ValueError(
            f'Unsupported path format: {val}. Available: gs, hail-az, local'
        )


class GSPathScheme(PathScheme):
    """
    Google Cloud Storage path scheme.
    """

    def __init__(self):
        self.scheme = 'gs'
        self.prefix = 'cpg'

    def path_prefix(self, dataset: str, category: str) -> str:
        """Build path prefix used in dataset_path"""
        return f'{self.prefix}-{dataset}-{category}'

    def full_path(self, prefix: str, suffix: str) -> str:
        """Build full path from prefix and suffix"""
        return os.path.join(f'{self.scheme}://', prefix, suffix)


class AzurePathScheme(PathScheme):
    """
    Azure Blob Storage path scheme, following the Hail Batch hail-az format.
    """

    def __init__(self, account: Optional[str] = 'cpg'):
        config = get_config()
        self.scheme = 'hail-az'
        self.account = config['workflow'].get('azure_account', account)

    def path_prefix(self, dataset: str, category: str) -> str:
        """Build path prefix used in dataset_path"""
        return f'{self.account}/{dataset}-{category}'

    def full_path(self, prefix: str, suffix: str) -> str:
        """Build full path from prefix and suffix"""
        return os.path.join(f'{self.scheme}://', prefix, suffix)


class LocalPathScheme(PathScheme):
    """
    Local posix path scheme. Requires workflow/local_dir to be set
    """

    def __init__(self):
        if not (local_dir := get_config()['workflow'].get('local_dir')):
            local_dir = tempfile.mkdtemp(prefix='cpg-utils-')
        self.local_dir = to_path(local_dir)
        self.scheme = 'local'

    def path_prefix(self, dataset: str, category: str) -> str:
        """Build path prefix used in dataset_path"""
        return f'{dataset}-{category}'

    def full_path(self, prefix: str, suffix: str) -> str:
        """Build full path from prefix and suffix"""
        return str(self.local_dir / prefix / suffix)


class Namespace(Enum):
    """
    Storage namespace.
    https://github.com/populationgenomics/team-docs/tree/main/storage_policies#main-vs-test
    """

    MAIN = 'main'
    TEST = 'test'

    @staticmethod
    def from_access_level(str_val: str) -> 'Namespace':
        """
        Parse value from an access level string.
        >>> Namespace.from_access_level('test')
        Namespace.TEST
        >>> Namespace.from_access_level('standard')
        Namespace.MAIN
        >>> Namespace.from_access_level('main')
        Namespace.MAIN
        """
        for val, str_vals in {
            Namespace.MAIN: ['main', 'standard', 'full'],
            Namespace.TEST: ['test'],
        }.items():
            if str_val in str_vals:
                return val
        raise ValueError(f'Cannot parse namespace or access level {str_val}')


def dataset_path(
    suffix: str,
    category: Optional[str] = None,
    dataset: Optional[str] = None,
    access_level: Optional[str] = None,
    path_scheme: Optional[str] = None,
) -> str:
    """
    Returns a full path for the current dataset, given a category and path suffix.

    This is useful for specifying input files, as in contrast to the output_path
    function, dataset_path does _not_ take the `workflow/output_prefix` config variable
    into account.

    Examples
    --------
    Assuming that the analysis-runner has been invoked with
    `--dataset fewgenomes --access-level test --output 1kg_pca/v42`:

    >>> from cpg_utils.hail_batch import dataset_path
    >>> dataset_path('1kg_densified/combined.mt')
    'gs://cpg-fewgenomes-test/1kg_densified/combined.mt'
    >>> dataset_path('1kg_densified/report.html', 'web')
    'gs://cpg-fewgenomes-test-web/1kg_densified/report.html'
    >>> dataset_path('1kg_densified/report.html', path_scheme='hail-az')
    'hail-az://cpg/fewgenomes-test/1kg_densified/report.html'

    Notes
    -----
    Requires either the
    * `workflow/dataset` and `workflow/access_level` config variables, or the
    * `workflow/dataset_path` config variable
    to be set, where the former takes precedence.

    Parameters
    ----------
    suffix : str
        A path suffix to append to the bucket.
    category : str, optional
        A category like "upload", "tmp", "web". If omitted, defaults to the "main" and
        "test" buckets based on the access level. See
        https://github.com/populationgenomics/team-docs/tree/main/storage_policies
        for a full list of categories and their use cases.
    dataset : str, optional
        Dataset name, takes precedence over the `workflow/dataset` config variable
    access_level : str, optional
        Access level, takes precedence over the `workflow/access_level` config variable
    path_scheme: str, optional
        Cloud storage path scheme, takes precedence over the `workflow/path_scheme`
        config variable

    Returns
    -------
    str
    """
    config = get_config()
    dataset = dataset or config['workflow'].get('dataset')
    access_level = access_level or config['workflow'].get('access_level')
    path_scheme = path_scheme or config['workflow'].get('path_scheme', 'gs')

    if dataset and access_level:
        namespace = Namespace.from_access_level(access_level)
        if category is None:
            category = namespace.value
        elif category != 'archive':
            category = f'{namespace.value}-{category}'
        prefix = PathScheme.parse(path_scheme).path_prefix(dataset, category)
    else:
        prefix = config['workflow']['dataset_path']

    return PathScheme.parse(path_scheme).full_path(prefix, suffix)


def web_url(
    suffix: str = '',
    dataset: Optional[str] = None,
    access_level: Optional[str] = None,
) -> str:
    """Returns URL corresponding to a dataset path of category 'web',
    assuming other arguments are the same.
    """
    config = get_config()
    dataset = dataset or config['workflow'].get('dataset')
    access_level = access_level or config['workflow'].get('access_level')
    namespace = Namespace.from_access_level(access_level)
    web_url_template = config['workflow'].get('web_url_template')
    try:
        url = web_url_template.format(dataset=dataset, namespace=namespace.value)
    except KeyError as e:
        raise ValueError(
            f'`workflow/web_url_template` should be parametrised by "dataset" and '
            f'"namespace" in curly braces, for example: '
            f'https://{{namespace}}-web.populationgenomics.org.au/{{dataset}}. '
            f'Got: {web_url_template}'
        ) from e
    return os.path.join(url, suffix)


def output_path(suffix: str, category: Optional[str] = None) -> str:
    """Returns a full path for the given category and path suffix.

    In contrast to the dataset_path function, output_path takes the `workflow/output_prefix`
    config variable into account.

    Examples
    --------
    If using the analysis-runner, the `workflow/output_prefix` would be set to the argument
    provided using the --output argument, e.g.
    `--dataset fewgenomes --access-level test --output 1kg_pca/v42`:
    will use '1kg_pca/v42' as the base path to build upon in this method

    >>> from cpg_utils.hail_batch import output_path
    >>> output_path('loadings.ht')
    'gs://cpg-fewgenomes-test/1kg_pca/v42/loadings.ht'
    >>> output_path('report.html', 'web')
    'gs://cpg-fewgenomes-test-web/1kg_pca/v42/report.html'

    Notes
    -----
    Requires the `workflow/output_prefix` config variable to be set, in addition to the
    requirements for `dataset_path`.

    Parameters
    ----------
    suffix : str
        A path suffix to append to the bucket + output directory.
    category : str, optional
        A category like "upload", "tmp", "web". If omitted, defaults to the "main" and
        "test" buckets based on the access level. See
        https://github.com/populationgenomics/team-docs/tree/main/storage_policies
        for a full list of categories and their use cases.

    Returns
    -------
    str
    """
    return dataset_path(
        os.path.join(get_config()['workflow']['output_prefix'], suffix), category
    )


def image_path(key: str) -> str:
    """Returns a path to a container image in the default registry using the
    key in the config's images section.

    Examples
    --------
    >>> image_path('bcftools')
    'australia-southeast1-docker.pkg.dev/cpg-common/images/bcftools:1.10.2'

    Notes
    -----
    Requires config variables `workflow/image_registry_prefix` and `images/<key>`.

    Parameters
    ----------
    key : str
        Describes the key within the `images` config section.

    Returns
    -------
    str
    """
    suffix = get_config()['images'][key]
    return os.path.join(get_config()['workflow']['image_registry_prefix'], suffix)


def reference_path(key: str) -> Path:
    """Returns a path to a file in the references bucket using the key in
    the config's references section.

    Examples
    --------
    >>> reference_path('vep_mount')
    CloudPath('gs://cpg-reference/vep/105.0/mount')
    >>> reference_path('broad/genome_calling_interval_lists')
    CloudPath('gs://cpg-reference/hg38/v0/wgs_calling_regions.hg38.interval_list')

    Notes
    -----
    Requires the `workflow/reference_prefix` config variable to be set.

    Parameters
    ----------
    key : str
        Describes the key within the `references` config section. Can specify
        nested sections with a "/" separator.

    Returns
    -------
    str
    """
    prefix = to_path(get_config()['workflow']['reference_prefix'])
    d = get_config()['references']
    sections = key.strip('/').split('/')
    for section in sections[:-1]:
        if section not in d:
            raise ValueError(f'No subsection {section} in {str(d)}')
        d = d[section]
        if extra_prefix := d.get('prefix'):
            prefix /= extra_prefix
    suffix = d[sections[-1]]
    return prefix / suffix


def genome_build() -> str:
    """Return the genome build name"""
    return get_config()['references'].get('genome_build', 'GRCh38')


def fasta_res_group(b, indices: Optional[List] = None):
    """
    Hail Batch resource group for fasta reference files.
    @param b: Hail Batch object.
    @param indices: list of extensions to add to the base fasta file path.
    """
    ref_fasta = reference_path('broad/ref_fasta')
    d = dict(
        base=str(ref_fasta),
        fai=str(ref_fasta) + '.fai',
        dict=str(ref_fasta.with_suffix('.dict')),
    )
    if indices:
        for ext in indices:
            d[ext] = f'{ref_fasta}.{ext}'
    return b.read_input_group(**d)


def authenticate_cloud_credentials_in_job(
    job,
    print_all_statements: bool = True,
):
    """
    Takes a hail batch job, activates the appropriate service account

    Once multiple environments are supported this method will decide
    on which authentication method is appropriate

    Parameters
    ----------
    job
        * A hail BashJob
    print_all_statements
        * logging toggle

    Returns
    -------
    None
    """

    # Use "set -x" to print the commands for easier debugging.
    if print_all_statements:
        job.command('set -x')

    # activate the google service account
    job.command(GCLOUD_AUTH_COMMAND)


# commands that declare functions that pull files on an instance,
# handling transitive errors
RETRY_CMD = """\
function fail {
  echo $1 >&2
  exit 1
}

function retry {
  local n_attempts=10
  local delay=30
  local n=1
  while ! eval "$@"; do
    if [[ $n -lt $n_attempts ]]; then
      ((n++))
      echo "Command failed. Attempt $n/$n_attempts after ${delay}s..."
      sleep $delay;
    else
      fail "The command has failed after $n attempts."
    fi
  done
}

function retry_gs_cp {
  src=$1

  if [ -n "$2" ]; then
    dst=$2
  else
    dst=/io/batch/${basename $src}
  fi

  retry gsutil -o GSUtil:check_hashes=never cp $src $dst
}
"""

# command that monitors the instance storage space
MONITOR_SPACE_CMD = f'df -h; du -sh /io; du -sh /io/batch'

ADD_SCRIPT_CMD = """\
cat <<EOT >> {script_name}
{script_contents}
EOT\
"""


def command(
    cmd: Union[str, List[str]],
    monitor_space: bool = False,
    setup_gcp: bool = False,
    define_retry_function: bool = False,
    rm_leading_space: bool = True,
    python_script_path: Optional[Path] = None,
) -> str:
    """
    Wraps a command for Batch.

    @param cmd: command to wrap (can be a list of commands)
    @param monitor_space: add a background process that checks the instance disk
        space every 5 minutes and prints it to the screen
    @param setup_gcp: authenticate on GCP
    @param define_retry_function: when set, adds bash functions `retry` that attempts
        to redo a command after every 30 seconds (useful to pull inputs
        and get around GoogleEgressBandwidth Quota or other google quotas)
    @param rm_leading_space: remove all leading spaces and tabs from the command lines
    @param python_script_path: if provided, copy this python script into the command
    """
    if isinstance(cmd, list):
        cmd = '\n'.join(cmd)

    cmd = f"""\
    set -o pipefail
    set -ex
    {GCLOUD_AUTH_COMMAND if setup_gcp else ''}
    {RETRY_CMD if define_retry_function else ''}

    {f'(while true; do {MONITOR_SPACE_CMD}; sleep 600; done) &'
    if monitor_space else ''}

    {{copy_script_cmd}}

    {cmd}

    {MONITOR_SPACE_CMD if monitor_space else ''}
    """

    if rm_leading_space:
        # remove any leading spaces and tabs
        cmd = '\n'.join(line.strip() for line in cmd.split('\n'))
        # remove stretches of spaces
        cmd = '\n'.join(' '.join(line.split()) for line in cmd.split('\n'))
    else:
        # Remove only common leading space:
        cmd = textwrap.dedent(cmd)

    # We don't want the python script tabs to be stripped, so
    # we are inserting it after leading space is removed
    if python_script_path:
        with python_script_path.open() as f:
            script_contents = f.read()
        cmd = cmd.replace(
            '{copy_script_cmd}',
            ADD_SCRIPT_CMD.format(
                script_name=python_script_path.name,
                script_contents=script_contents,
            ),
        )
    else:
        cmd = cmd.replace('{copy_script_cmd}', '')

    return cmd


def query_command(
    module,
    func_name: str,
    *func_args,
    setup_gcp: bool = False,
    setup_hail: bool = True,
    packages: Optional[List[str]] = None,
) -> str:
    """
    Construct a command to run a python function inside a Hail Batch job.
    If hail_billing_project is provided, Hail Query will be also initialised.

    Run a Python Hail Query function inside a Hail Batch job.
    Constructs a command string to use with job.command().
    If hail_billing_project is provided, Hail Query will be initialised.
    """
    init_hail_code = """
from cpg_utils.hail_batch import init_batch
init_batch()
"""

    python_code = f"""
{'' if not setup_hail else init_hail_code}
{inspect.getsource(module)}
{func_name}{func_args}
"""

    return f"""\
set -o pipefail
set -ex
{GCLOUD_AUTH_COMMAND if setup_gcp else ''}

{('pip3 install ' + ' '.join(packages)) if packages else ''}

cat << EOT >> script.py
{python_code}
EOT
python3 script.py
"""



"""
Sending notifications to Slack. To enable, create a channel, add "Seqr Loader" app
into a channel with:

/invite @Seqr Loader

Make sure `slack/channel`, `slack/token_secret_id`, and `slack/token_project_id`
configuration values are set.
"""

from textwrap import dedent

from cpg_utils.cloud import read_secret
from cpg_utils.config import get_config

from hailtop.batch.job import Job


def get_token() -> str:
    """
    Returns Slack token.
    """
    token_secret_id = get_config()['slack'].get('token_secret_id')
    token_project_id = get_config()['slack'].get('token_project_id')
    if not token_secret_id or not token_project_id:
        raise ValueError(
            '`slack.token_secret_id` and `slack.token_project_id` '
            'must be set in config to retrieve Slack token'
        )
    token = read_secret(
        project_id=token_project_id,
        secret_name=token_secret_id,
        fail_gracefully=False,
    )
    assert token
    return token


def slack_env(j: Job):
    """
    Add environment variables that configure Slack reporter.
    """
    if not (channel := get_config().get('slack', {}).get('channel')):
        return None
    j.env('SLACK_CHANNEL', channel)
    j.env('SLACK_TOKEN', get_token())


def slack_message_cmd(
    j: Job,
    text: str | None = None,
    data: dict[str, str] | None = None,
):
    """
    Make a bash command that prepares and sends a Slack message.
    Message can be constructed from a dict `data`, or can be passed directly
    as text (`text`). In either case, strings can use Slack `mrkdwn`
    for formatting: https://api.slack.com/reference/surfaces/formatting

    Assumes `slack/channel` configuration parameter is set, as well as
    SLACK_TOKEN environment variable.
    """
    if not (channel := get_config()['slack'].get('channel')):
        return ''
    msg = ''
    if data:
        msg += '\\n'.join(f'{k}: {v}' for k, v in data.items())
    if text:
        msg += f'\\n{text}'
    if not msg:
        return ''
    slack_env(j)
    j.command(
        dedent(
            f"""
    curl -X POST \
    -H "Authorization: Bearer {get_token()}" \
    -H "Content-type: application/json" \
    -d '{{"channel": "'{channel}'", "text": "{msg}"}}' \
    https://slack.com/api/chat.postMessage
    """
        )
    )



"""
Extend cloudpathlib Azure implementation to support hail-az:// scheme.
Inspired by https://github.com/drivendataorg/cloudpathlib/issues/157
"""

import re
import os
import json
import logging
from typing import Union, Optional, Any, Callable
from urllib.parse import urlparse
import mimetypes
from azure.identity import DefaultAzureCredential, ClientSecretCredential
from azure.storage.blob import BlobServiceClient
#import azure.storage.blob
from cloudpathlib import AzureBlobClient, AzureBlobPath
from cloudpathlib.client import register_client_class
from cloudpathlib.cloudpath import register_path_class, CloudPath
from cloudpathlib.exceptions import InvalidPrefixError

# TODO pretty heavy handed
logging.getLogger('azure').setLevel(logging.WARN)
logging.getLogger('msal').setLevel(logging.WARN)
logging.getLogger('urllib3').setLevel(logging.WARN)

@register_client_class('hail-az')
class HailAzureBlobClient(AzureBlobClient):
    def __init__(
        self,
        account_url: str,
        local_cache_dir: Optional[Union[str, os.PathLike]] = None,
        content_type_method: Optional[Callable] = mimetypes.guess_type,
    ):
        """Class constructor. Sets up a [`BlobServiceClient`](
        https://docs.microsoft.com/en-us/python/api/azure-storage-blob/azure.storage.blob.blobserviceclient?view=azure-python).
        Supports the following authentication methods, specific to Hail Batch.
        - Implicit instantiation of a [`DefaultAzureCredential`](https://learn.microsoft.com/en-us/dotnet/api/azure.identity.defaultazurecredential?view=azure-dotnet-preview)
        - Environment variable `"AZURE_APPLICATION_CREDENTIALS"` containing a path to a JSON file that in turn contains three fields: 
        "tenant", "appId", and "password".
        - Account URL via `account_url`, authenticated with an embedded SAS token. In all cases an Account URL is required, and if a SAS token is provided, it will be used for Authentication.
        If multiple methods are used, priority order is reverse of list above (later in list takes
        priority).
        Args:
            account_url (str): The URL to the blob storage account, optionally
                authenticated with a SAS token. See documentation for [`BlobServiceClient`](
                https://docs.microsoft.com/en-us/python/api/azure-storage-blob/azure.storage.blob.blobserviceclient?view=azure-python).
            local_cache_dir (Optional[Union[str, os.PathLike]]): Path to directory to use as cache
                for downloaded files. If None, will use a temporary directory.
            content_type_method (Optional[Callable]): Function to call to guess media type (mimetype) when
                writing a file to the cloud. Defaults to `mimetypes.guess_type`. Must return a tuple (content type, content encoding).
        """
        if account_url is None:
            raise ValueError("account_url must be specified")

        parsed = urlparse(account_url)
        if bool(parsed.query):
            # Use passed embedded SAS token
            service_client = BlobServiceClient(account_url=account_url)
        elif (azure_application_credentials_file := os.getenv("AZURE_APPLICATION_CREDENTIALS")) is not None:
            msal_credential = self._msal_credential_from_file(azure_application_credentials_file)
            service_client = BlobServiceClient(account_url=account_url, credential=msal_credential)
        else:
            # EnvironmentCredential, ManagedIdentityCredential, AzureCliCredential
            msal_credential = DefaultAzureCredential(
                exclude_powershell_credential = True,
                exclude_visual_studio_code_credential = True,
                exclude_shared_token_cache_credential = True,
                exclude_interactive_browser_credential = True
            )
            service_client = BlobServiceClient(account_url=account_url, credential=msal_credential)

        super().__init__(blob_service_client=service_client, local_cache_dir=local_cache_dir, content_type_method=content_type_method)

    def _msal_credential_from_file(self, file_path: str) -> ClientSecretCredential:
        with open(file_path, "r") as f:
            credentials = json.loads(f.read())
        return ClientSecretCredential(
            tenant_id=credentials["tenant"],
            client_id=credentials["appId"],
            client_secret=credentials["password"]
        )


@register_path_class('hail-az')
class HailAzureBlobPath(AzureBlobPath):
    """
    Extending Path implementation to support hail-az:// and https:// schemes
    >>> CloudPath('hail-az://myaccount/mycontainer/tmp')
    HailAzureBlobPath('hail-az://myaccount/mycontainer/tmp')
    >>> CloudPath('https://myaccount.blob.core.windows.net/mycontainer/tmp')
    HailAzureBlobPath('hail-az://myaccount/mycontainer/tmp')
    """

    cloud_prefix: str = 'hail-az://'
    client: 'HailAzureBlobClient'

    def __init__(
        self,
        cloud_path: Union[str, CloudPath],
        client: Optional[HailAzureBlobClient] = None,
    ):
        if isinstance(cloud_path, str):
            parsed = urlparse(cloud_path)
            m = re.match(
                r'(?P<account>[a-z0-9]+)(\.(?P<type>blob|dfs)(\.core\.windows\.net)?)?',
                parsed.netloc,
                flags=re.IGNORECASE,
            )
            if m is None or not self.is_valid_cloudpath(cloud_path):
                raise ValueError(f'Bad Azure path "{cloud_path}"')
            account = m.group('account')
            fstype = m.group('type') or 'blob'
            account_url = f'https://{account}.{fstype}.core.windows.net/'
            cloud_path = (
                f'{HailAzureBlobPath.cloud_prefix}{account}/'
                f'{parsed.path.lstrip("/")}'
            )
            if client is None:
                if parsed.query:
                    account_url = account_url + "?" + parsed.query
                client = HailAzureBlobClient(account_url=account_url)

        super().__init__(cloud_path, client=client)

    @classmethod
    def is_valid_cloudpath(
        cls, path: Union[str, CloudPath], raise_on_error=False
    ) -> bool:
        """
        Also allowing HTTP.
        """
        valid = bool(
            re.match(
                fr'({HailAzureBlobPath.cloud_prefix}[a-z0-9]+|https://[a-z0-9]+\.(blob|dfs)\.core\.windows\.net)/[a-z0-9]+',
                str(path).lower(),
            )
        )
#                fr'({HailAzureBlobPath.cloud_prefix}|https://[a-z0-9]+\.(blob|dfs)\.core\.windows\.net)/',

        if raise_on_error and not valid:
            raise InvalidPrefixError(
                f'{path} is not a valid path since it does not start with {cls.cloud_prefix} '
                f'or valid Azure https blob or dfs location.'
            )

        return valid

    @property
    def account(self) -> str:
        """
        Just the account part.
        """
        return self._no_prefix.split('/', 2)[0]

    @property
    def container(self) -> str:
        """
        Minus the account part.
        """
        return self._no_prefix.split('/', 2)[1]

    @property
    def blob(self) -> str:
        """
        No prefix, no account part.
        """
        parts = self._no_prefix.split('/', 2)
        return parts[2] if len(parts) == 3 else None



"""
Functions to set up Hail Batch resources (cores, memory, storage).
"""

import math
from dataclasses import dataclass

from cpg_utils.config import get_config
from hailtop.batch.job import Job


def _is_power_of_two(n: int) -> bool:
    return math.ceil(math.log(n, 2)) == math.floor(math.log(n, 2))


def gcp_machine_name(name: str, ncpu: int) -> str:
    """
    Machine type name in the GCP world
    """
    assert name in ['standard', 'highmem', 'highcpu'], name
    assert _is_power_of_two(ncpu), ncpu
    return f'n1-{name}-{ncpu}'


@dataclass(init=False)
class MachineType:
    """
    Hail Batch machine type on GCP
    """

    min_cpu: int = 2
    threads_on_cpu = 2  # hyperthreading

    def __init__(
        self,
        name: str,
        ncpu: int,
        mem_gb_per_core: float,
        price_per_hour: float,
    ):
        self.name = name
        self.max_ncpu = ncpu
        self.mem_gb_per_core = mem_gb_per_core
        self.price_per_hour = price_per_hour

    def gcp_name(self):
        """
        Machine type name in the GCP world
        """
        return gcp_machine_name(self.name, self.max_ncpu)

    def max_threads(self) -> int:
        """
        Number of available threads
        """
        return self.max_ncpu * self.threads_on_cpu

    def calc_instance_disk_gb(self) -> int:
        """
        The maximum available storage on an instance is calculated
        in `batch/batch/utils.py/unreserved_worker_data_disk_size_gib()`
        as the disk size (375G) minus reserved image size (30G) minus
        reserved storage per core (5G*ncpu = 120G for a 32-core instance),
        """
        disk_size_gb = 375
        reserved_gb = 30
        reserved_gb_per_core = 5
        return disk_size_gb - reserved_gb - reserved_gb_per_core * self.max_ncpu

    def set_resources(
        self,
        j: Job,
        fraction: float | None = None,
        ncpu: int | None = None,
        nthreads: int | None = None,
        mem_gb: float | None = None,
        storage_gb: float | None = None,
    ) -> 'JobResource':
        """
        Set resources to a Job object. If any optional parameters are set,
        they will be used as a bound to request a fraction of an instance.
        """
        return self.request_resources(
            fraction=fraction,
            ncpu=ncpu,
            nthreads=nthreads,
            mem_gb=mem_gb,
            storage_gb=storage_gb,
        ).set_to_job(j)

    def request_resources(
        self,
        fraction: float | None = None,
        ncpu: int | None = None,
        nthreads: int | None = None,
        mem_gb: float | None = None,
        storage_gb: float | None = None,
    ) -> 'JobResource':
        """
        Request resources from the machine, satisfying all provided requirements.
        If not requirements are provided, the minimal amount of cores
        (self.MIN_NCPU) will be used.
        """
        # determining the biggest limit to satisfy, measured in the number of CPUs:
        min_ncpu = max(
            filter(
                None,
                [
                    self.adjust_ncpu(ncpu or self.min_cpu),
                    self.fraction_to_ncpu(fraction) if fraction else None,
                    self.nthreads_to_ncpu(nthreads) if nthreads else None,
                    self.mem_gb_to_ncpu(mem_gb) if mem_gb else None,
                    self.storage_gb_to_ncpu(storage_gb) if storage_gb else None,
                ],
            )
        )
        return JobResource(
            machine_type=self,
            ncpu=min_ncpu,
            attach_disk_storage_gb=(
                storage_gb
                if storage_gb and storage_gb > self.calc_instance_disk_gb()
                else None
            ),
        )

    def fraction_to_ncpu(self, fraction: float) -> int:
        """
        Converts fraction to the number of CPU (e.g. fraction=1.0 to take the entire
        machine, fraction=0.5 to take half of it, etc.).
        """
        ncpu = int(math.ceil(self.max_ncpu * fraction))
        return self.adjust_ncpu(ncpu)

    def mem_gb_to_ncpu(self, mem_gb: float) -> int:
        """
        Converts memory requirement to the number of CPU requirement.
        """
        ncpu = int(math.ceil(mem_gb / self.mem_gb_per_core))
        return self.adjust_ncpu(ncpu)

    def storage_gb_to_ncpu(self, storage_gb: float) -> int:
        """
        Converts storage requirement to the number of CPU requirement.

        We want to avoid attaching disks: attaching a disk to an existing instance
        might fail with `mkfs.ext4 ...` error, see:
        https://batch.hail.populationgenomics.org.au/batches/7488/jobs/12
        So this function will calculate the number of CPU to request so your jobs
        can be packed to fit the default instance's available storage
        (caluclated with self.calc_instance_disk_gb()).
        """
        fraction = storage_gb / self.calc_instance_disk_gb()
        fraction = min(fraction, 1.0)
        return self.fraction_to_ncpu(fraction)

    def nthreads_to_ncpu(self, nthreads: int) -> int:
        """
        Convert number of threads into number of cores/CPU
        """
        return self.adjust_ncpu(math.ceil(nthreads / 2))

    def adjust_ncpu(self, ncpu: int) -> int:
        """
        Adjust request number of CPU to a correct requestable number:
        the nearest power of 2, not less than the minimal number of cores allowed.
        """
        if ncpu > self.max_ncpu:
            ValueError(
                f'Requesting more cores than available on {self.name} machine: '
                f'{ncpu}>{self.max_ncpu}'
            )

        if ncpu < MachineType.min_cpu:
            ncpu = MachineType.min_cpu

        # round to the nearest power of 2 (15 -> 16, 16 -> 16, 17 -> 32)
        return int(pow(2, math.ceil(math.log2(ncpu))))


# Default Hail Batch machine.
#
# Bigger default number of cores (32 vs 16) would allow for more threads for alignment,
# however would mean a smaller available storage (5G per core is reserved)
STANDARD = MachineType(
    'standard',
    ncpu=16,  # 32
    mem_gb_per_core=3.75,  # Total 60G
    price_per_hour=1.0787,  # 32-core machine would cost 2.1574
)

# ~1.7 times memory per core than standard-32, but only 16 cores.
# Total 104G - smaller than standard-32, but less expensive, so useful for
# memory consuming tools that don't benefit from multiple threads
HIGHMEM = MachineType(
    'highmem',
    ncpu=16,
    mem_gb_per_core=6.5,
    price_per_hour=1.3431,
)


@dataclass(init=False)
class JobResource:
    """
    Represents a fraction of a Hail Batch instance.
    """

    def __init__(
        self,
        machine_type: MachineType,
        ncpu: int | None = None,
        attach_disk_storage_gb: float | None = None,
    ):
        """
        @param machine_type: Hail Batch machine pool type
        @param ncpu: number of CPU request. Will be used to calculate the fraction of
            the machine to take. If not set, all machine's CPUs will be used.
        @param attach_disk_storage_gb: if set to > MachineType.max_default_storage_gb,
            a larger disc will be attached by Hail Batch.
        """
        self.machine_type = machine_type

        self.fraction_of_full: float = 1.0
        if ncpu is not None:
            if ncpu > self.machine_type.max_ncpu:
                raise ValueError(
                    f'Max number of CPU on machine {self.machine_type.name} '
                    f'is {self.machine_type.max_ncpu}, requested {ncpu}'
                )
            self.fraction_of_full = ncpu / self.machine_type.max_ncpu

        self.attach_disk_storage_gb = None
        if attach_disk_storage_gb is not None:
            if self.fraction_of_full < 1:
                raise ValueError(
                    f'Storage can be overridden only when the entire machine is used, '
                    f'not a fraction ({self.fraction_of_full}). '
                    f'override_storage_gb={attach_disk_storage_gb}'
                )
            self.attach_disk_storage_gb = attach_disk_storage_gb

    def get_mem_gb(self) -> float:
        """
        Memory resources in GB
        """
        return self.get_ncpu() * self.machine_type.mem_gb_per_core

    def get_java_mem_mb(self) -> int:
        """
        Calculate memory to pass to the `java -Xms` option.
        Subtracts 1G to start a java VM, and converts to MB as the option doesn't
        support fractions of GB.
        """
        return int(math.floor((self.get_mem_gb() - 1) * 1000))

    def get_ncpu(self) -> int:
        """
        Number of cores/CPU
        """
        return int(self.machine_type.max_ncpu * self.fraction_of_full)

    def get_nthreads(self) -> int:
        """
        Number of threads
        """
        return self.get_ncpu() * MachineType.threads_on_cpu

    def get_storage_gb(self) -> float:
        """
        Calculate storage in GB
        """
        if self.attach_disk_storage_gb:
            storage_gb = self.attach_disk_storage_gb
        else:
            storage_gb = (
                self.machine_type.calc_instance_disk_gb() * self.fraction_of_full
            )

        # Hail Batch actually requests 5% lower number than the
        # requested one (e.g. "req_storage: 46.25G, actual_storage: 44.0 GiB"),
        # so we will ask for a bigger number.
        return storage_gb * 1.05

    def set_to_job(self, j: Job) -> 'JobResource':
        """
        Set the resources to a Job object. Return self to allow chaining, e.g.:
        >>> nthreads = STANDARD.request_resources(nthreads=4).set_to_job(j).get_nthreads()
        """

        j.storage(f'{self.get_storage_gb()}G')
        j.cpu(self.get_ncpu())
        j.memory(f'{self.get_mem_gb()}G')

        # Returning self to allow command chaining.
        return self


def storage_for_cram_qc_job() -> int | None:
    """
    Get storage request for a CRAM QC processing job, gb
    """
    sequencing_type = get_config()['workflow']['sequencing_type']
    storage_gb = None  # avoid extra disk by default
    if sequencing_type == 'genome':
        storage_gb = 100
    if sequencing_type == 'exome':
        storage_gb = 20
    return storage_gb



"""
Utility functions and constants.
"""

import logging
import re
import string
import sys
import time
import traceback
import unicodedata
from functools import lru_cache
from random import choices
from typing import cast

from hailtop.batch import ResourceFile

from cpg_utils import Path, to_path


@lru_cache
def exists(path: Path | str, verbose: bool = True) -> bool:
    """
    `exists_not_cached` that caches the result.

    The python code runtime happens entirely during the workflow construction,
    without waiting for it to finish, so there is no expectation that the object
    existence status would change during the runtime. This, this function uses
    `@lru_cache` to make sure that object existence is checked only once.
    """
    return exists_not_cached(path, verbose)


def exists_not_cached(path: Path | str, verbose: bool = True) -> bool:
    """
    Check if the object exists, where the object can be:
        * local file
        * local directory
        * cloud object
        * cloud URL representing a *.mt or *.ht Hail data,
          in which case it will check for the existence of a
          *.mt/_SUCCESS or *.ht/_SUCCESS file.
    @param path: path to the file/directory/object/mt/ht
    @param verbose: print on each check
    @return: True if the object exists
    """
    path = cast(Path, to_path(path))

    # rstrip to ".mt/" -> ".mt"
    if any(str(path).rstrip('/').endswith(f'.{suf}') for suf in ['mt', 'ht']):
        path = path / '_SUCCESS'

    if verbose:
        # noinspection PyBroadException
        try:
            res = path.exists()
        except BaseException:
            traceback.print_exc()
            logging.error(f'Failed checking {path}')
            sys.exit(1)
        logging.debug(f'Checked {path} [' + ('exists' if res else 'missing') + ']')
        return res
    return path.exists()


def can_reuse(
    path: list[Path] | Path | str | None,
    overwrite: bool,
) -> bool:
    """
    Checks if `fpath` is good to reuse in the analysis: it exists
    and `overwrite` is False.

    If `fpath` is a collection, it requires all files in it to exist.
    """
    if overwrite:
        return False

    if not path:
        return False

    if isinstance(path, list):
        return all(can_reuse(fp, overwrite) for fp in path)

    if not exists(path):
        return False

    logging.debug(f'Reusing existing {path}. Use --overwrite to overwrite')
    return True


def timestamp(rand_suffix_len: int = 5) -> str:
    """
    Generate a timestamp string. If `rand_suffix_len` is set, adds a short random
    string of this length for uniqueness.
    """
    result = time.strftime('%Y_%m%d_%H%M')
    if rand_suffix_len:
        rand_bit = ''.join(
            choices(string.ascii_uppercase + string.digits, k=rand_suffix_len)
        )
        result += f'_{rand_bit}'
    return result


def slugify(line: str):
    """
    Slugify a string.

    Example:
    >>> slugify(u"Héllø Wörld")
    u"hello-world"
    """

    return re.sub(
        r'[-\s]+',
        '-',
        re.sub(
            r'[^\w\s-]',
            '',
            unicodedata.normalize('NFKD', line).encode('ascii', 'ignore').decode(),
        )
        .strip()
        .lower(),
    )


def rich_sample_id_seds(
    rich_id_map: dict[str, str],
    file_names: list[str | ResourceFile],
) -> str:
    """
    Helper function to add seds into a command that would extend samples IDs
    in each file in `file_names` with an external ID, only if external ID is
    different from the original.

    @param rich_id_map: map used to replace samples, e.g. {'CPG1': 'CPG1|EXTID'}
    @param file_names: file names and Hail Batch Resource files where to replace IDs
    @return: bash command that does replacement
    """
    cmd = ''
    for sid, rich_sid in rich_id_map.items():
        for fname in file_names:
            cmd += f'sed -iBAK \'s/{sid}/{rich_sid}/g\' {fname}'
            cmd += '\n'
    return cmd




import json
from os import getenv

import azure.identity
import google.auth
import google.auth.exceptions
import google.auth.transport.requests


def get_azure_auth_token(scope: str) -> str:
    """Get Azure auth token for a particular scope"""
    return get_azure_credentials().get_token(scope).token


def get_azure_credentials() -> object:
    """
    Get Azure client credentials in one of two ways:
    - if AZURE_APPLICATION_CREDENTIALS is set, then grab from there
    - otherwise let the azure MSAL figure out the default 
    """
    if (credentials_filename := getenv("AZURE_APPLICATION_CREDENTIALS")):
        with open(credentials_filename, "r") as f:
            credentials = json.loads(f.read())
        credential = azure.identity.ClientSecretCredential(
            tenant_id=credentials["tenant"],
            client_id=credentials["appId"],
            client_secret=credentials["password"]
        )
    else:
        # EnvironmentCredential, ManagedIdentityCredential, AzureCliCredential
        credential = azure.identity.DefaultAzureCredential(
            exclude_powershell_credential = True,
            exclude_visual_studio_code_credential = True,
            exclude_shared_token_cache_credential = True,
            exclude_interactive_browser_credential = True
        )

    return credential


def get_google_auth_token(audience: str) -> str:
    """
    Get google auth token in one of two ways:
    - if GOOGLE_APPLICATION_CREDENTIALS is set, then grab from there
    - or run the equivalent of 'gcloud auth print-identity-token'
    ie: use service account identity token by default, then fallback otherwise
    https://stackoverflow.com/a/55804230
    """
    if (credentials_filename := getenv("GOOGLE_APPLICATION_CREDENTIALS")):
        with open(credentials_filename, "r") as f:
            from google.oauth2 import service_account

            info = json.load(f)
            credentials_content = (info.get("type") == "service_account") and info or None
            credentials = service_account.IDTokenCredentials.from_service_account_info(
                credentials_content, target_audience=audience
            )
            auth_req = google.auth.transport.requests.Request()
            credentials.refresh(auth_req)
            return credentials.token
    else:
        creds, _ = google.auth.default()
        creds.refresh(google.auth.transport.requests.Request())
        return creds.id_token


def get_analysis_runner_token() -> str:
    """Get analysis-runner Bearer auth token for Azure or GCP depending on deployment config."""
    from .deploy_config import get_deploy_config
    deploy_config = get_deploy_config()

    if deploy_config.cloud == "azure":
        scope = f"api://arapi-{deploy_config.analysis_runner_project}/.default"
        return get_azure_auth_token(scope)

    assert deploy_config.cloud == "gcp"
    audience = deploy_config.analysis_runner_host
    return get_google_auth_token(audience)


def get_sample_metadata_token() -> str:
    """Get sample-metadata Bearer auth token for Azure or GCP depending on deployment config."""
    from .deploy_config import get_deploy_config
    deploy_config = get_deploy_config()

    if deploy_config.cloud == "azure":
        scope = f"api://smapi-{deploy_config.sample_metadata_project}/.default"
        return get_azure_auth_token(scope)

    assert deploy_config.cloud == "gcp"
    audience = deploy_config.sample_metadata_host
    return get_google_auth_token(audience)



import os
import json

import pytest
from cpg_utils.deploy_config import (
    DEFAULT_CONFIG,
    DeployConfig,
    get_deploy_config,
    set_deploy_config,
    set_deploy_config_from_env
)


def test_default_config(monkeypatch):
    monkeypatch.delenv("CPG_DEPLOY_CONFIG", raising=False)
    monkeypatch.delenv("CLOUD", raising=False)
    set_deploy_config_from_env()
    dc = get_deploy_config()
    assert dc.to_dict() == DEFAULT_CONFIG


def test_env_config(monkeypatch, json_load):
    cfg1 = json_load("config_01.json")
    monkeypatch.setenv("CPG_DEPLOY_CONFIG", json.dumps(cfg1))
    set_deploy_config(DeployConfig.from_environment())
    dc = get_deploy_config()
    print(json.dumps(dc.to_dict(), indent=4))
    assert dc.to_dict() == cfg1


def test_env_override_config(monkeypatch):
    monkeypatch.delenv("CPG_DEPLOY_CONFIG", raising=False)
    monkeypatch.setenv("CLOUD", "bogus")
    with pytest.raises(AssertionError, match="Invalid cloud specification"):
        dc = DeployConfig.from_environment()
    monkeypatch.setenv("CLOUD", "azure")
    dc = DeployConfig.from_environment()
    assert dc.cloud == "azure"


def test_config_from_dict(json_load):
    cfg1 = json_load("config_01.json")
    dc = DeployConfig.from_dict(cfg1)
    assert dc.to_dict() == cfg1


def test_config_from_partial_dict(json_load):
    cfg1 = json_load("config_01.json")
    del cfg1["cloud"]
    with pytest.raises(TypeError, match="missing 1 required positional argument: 'cloud'"):
        dc = DeployConfig.from_dict(cfg1)



"""
Metamist wrapper to get input samples.
"""

import logging
from collections import defaultdict
from itertools import groupby

from cpg_utils.config import get_config

from .metamist import get_metamist, Sequence, AnalysisType, MetamistError
from .targets import Cohort, Sex, PedigreeInfo


_cohort: Cohort | None = None


def get_cohort() -> Cohort:
    """Return the cohort object"""
    global _cohort
    if not _cohort:
        _cohort = create_cohort()
    return _cohort


def create_cohort() -> Cohort:
    """
    Add datasets in the cohort. There exists only one cohort for the workflow run.
    """
    analysis_dataset_name = get_config()['workflow']['dataset']
    dataset_names = get_config()['workflow'].get('datasets', [analysis_dataset_name])
    skip_datasets = get_config()['workflow'].get('skip_datasets', [])
    dataset_names = [d for d in dataset_names if d not in skip_datasets]

    skip_samples = get_config()['workflow'].get('skip_samples', [])
    only_samples = get_config()['workflow'].get('only_samples', [])

    cohort = Cohort()
    for dataset_name in dataset_names:
        dataset = cohort.create_dataset(dataset_name)
        logging.info(f'Getting samples for dataset {dataset_name}')
        sample_entries = get_metamist().sapi.get_samples(
            body_get_samples={'project_ids': [dataset_name]}
        )
        sample_entries = _filter_samples(
            sample_entries,
            dataset_name,
            skip_samples,
            only_samples,
        )
        for entry in sample_entries:
            dataset.add_sample(
                id=str(entry['id']),
                external_id=str(entry['external_id']),
                meta=entry.get('meta', {}),
            )

    if not cohort.get_datasets():
        msg = 'No datasets populated'
        if skip_samples or only_samples or skip_datasets:
            msg += ' (after skipping/picking samples)'
        logging.warning(msg)
        return cohort

    if sequencing_type := get_config()['workflow'].get('sequencing_type'):
        _populate_alignment_inputs(cohort, sequencing_type)
        _filter_sequencing_type(cohort, sequencing_type)
    _populate_analysis(cohort)
    _populate_participants(cohort)
    _populate_pedigree(cohort)
    return cohort


def _filter_sequencing_type(cohort: Cohort, sequencing_type: str):
    """
    Filtering to the samples with only requested sequencing types.
    """
    for s in cohort.get_samples():
        if not s.seq_by_type:
            logging.warning(f'{s}: skipping because no sequencing inputs found')
            s.active = False
            continue

        if s.alignment_input_by_seq_type:
            avail_types = list(s.seq_by_type.keys())
            s.alignment_input_by_seq_type = {
                k: v
                for k, v in s.alignment_input_by_seq_type.items()
                if k == sequencing_type
            }
            if not bool(s.alignment_input_by_seq_type):
                logging.warning(
                    f'{s}: skipping because no inputs with data type '
                    f'"{sequencing_type}" found in {avail_types}'
                )
                s.active = False


def _filter_samples(
    entries: list[dict[str, str]],
    dataset_name: str,
    skip_samples: list[str] | None = None,
    only_samples: list[str] | None = None,
) -> list[dict]:
    """
    Apply the only_samples and skip_samples filters.
    """

    filtered_entries = []
    for entry in entries:
        cpgid = entry['id']
        extid = entry['external_id']
        if only_samples:
            if cpgid in only_samples or extid in only_samples:
                logging.info(f'Picking sample: {dataset_name}|{cpgid}|{extid}')
            else:
                continue
        if skip_samples:
            if cpgid in skip_samples or extid in skip_samples:
                logging.info(f'Skipping sample: {dataset_name}|{cpgid}|{extid}')
                continue
        filtered_entries.append(entry)
    return filtered_entries


def _populate_alignment_inputs(
    cohort: Cohort,
    sequencing_type: str,
    check_existence: bool = False,
) -> None:
    """
    Populate sequencing inputs for samples.
    """
    assert cohort.get_sample_ids()
    found_seqs: list[dict] = get_metamist().seqapi.get_sequences_by_sample_ids(
        cohort.get_sample_ids(), get_latest_sequence_only=False
    )
    found_seqs = [seq for seq in found_seqs if str(seq['type']) == sequencing_type]
    found_seqs_by_sid = defaultdict(list)
    for found_seq in found_seqs:
        found_seqs_by_sid[found_seq['sample_id']].append(found_seq)

    # Log sequences without samples, this is a pretty common thing,
    # but useful to log to easier track down samples not processed
    if sample_wo_seq := [
        s for s in cohort.get_samples() if s.id not in found_seqs_by_sid
    ]:
        msg = f'No {sequencing_type} sequencing data found for samples:\n'
        ds_sample_count = {
            ds_name: len(list(ds_samples))
            for ds_name, ds_samples in groupby(
                cohort.get_samples(), key=lambda s: s.dataset.name
            )
        }
        for ds, samples in groupby(sample_wo_seq, key=lambda s: s.dataset.name):
            msg += (
                f'\t{ds}, {len(list(samples))}/{ds_sample_count.get(ds)} samples: '
                f'{", ".join([s.id for s in samples])}\n'
            )
        logging.info(msg)

    for sample in cohort.get_samples():
        for d in found_seqs_by_sid.get(sample.id, []):
            seq = Sequence.parse(d, check_existence=check_existence)
            sample.seq_by_type[seq.sequencing_type] = seq
            if seq.alignment_input:
                if seq.sequencing_type in sample.alignment_input_by_seq_type:
                    raise MetamistError(
                        f'{sample}: found more than 1 alignment input with '
                        f'sequencing type: {seq.sequencing_type}. Check your '
                        f'input provider to make sure there is only one data source '
                        f'of sequencing type per sample.'
                    )
                sample.alignment_input_by_seq_type[
                    seq.sequencing_type
                ] = seq.alignment_input


def _populate_analysis(cohort: Cohort) -> None:
    """
    Populate Analysis entries.
    """
    for dataset in cohort.get_datasets():
        gvcf_by_sid = get_metamist().find_analyses_by_sid(
            dataset.get_sample_ids(),
            analysis_type=AnalysisType.GVCF,
            dataset=dataset.name,
        )
        cram_by_sid = get_metamist().find_analyses_by_sid(
            dataset.get_sample_ids(),
            analysis_type=AnalysisType.CRAM,
            dataset=dataset.name,
        )
        for sample in dataset.get_samples():
            if (analysis := gvcf_by_sid.get(sample.id)) and analysis.output:
                assert analysis.output == sample.make_gvcf_path().path, (
                    analysis.output,
                    sample.make_gvcf_path().path,
                )
            if (analysis := cram_by_sid.get(sample.id)) and analysis.output:
                assert analysis.output == sample.make_cram_path().path, analysis.output


def _populate_participants(cohort: Cohort) -> None:
    """
    Populate Participant entries.
    """
    for dataset in cohort.get_datasets():
        logging.info(f'Reading participants IDs for dataset {dataset}')
        pid_sid_multi = (
            get_metamist().papi.get_external_participant_id_to_internal_sample_id(
                dataset.name
            )
        )
        participant_by_sid = {}
        for group in pid_sid_multi:
            pid = group[0]
            for sid in group[1:]:
                participant_by_sid[sid] = pid.strip()

        for sample in dataset.get_samples():
            if pid := participant_by_sid.get(sample.id):
                sample.participant_id = pid


def _populate_pedigree(cohort: Cohort) -> None:
    """
    Populate pedigree data for samples.
    """
    sample_by_participant_id = dict()
    for s in cohort.get_samples():
        sample_by_participant_id[s.participant_id] = s

    for dataset in cohort.get_datasets():
        logging.info(f'Reading pedigree for dataset {dataset}')
        ped_entries = get_metamist().get_ped_entries(dataset=dataset.name)
        ped_entry_by_participant_id = {}
        for ped_entry in ped_entries:
            part_id = str(ped_entry['individual_id'])
            ped_entry_by_participant_id[part_id] = ped_entry

        for sample in dataset.get_samples():
            if sample.participant_id not in ped_entry_by_participant_id:
                logging.warning(
                    f'No pedigree data for participant {sample.participant_id}'
                )
                continue

            ped_entry = ped_entry_by_participant_id[sample.participant_id]
            maternal_sample = sample_by_participant_id.get(
                str(ped_entry['maternal_id'])
            )
            paternal_sample = sample_by_participant_id.get(
                str(ped_entry['paternal_id'])
            )
            sample.pedigree = PedigreeInfo(
                sample=sample,
                fam_id=ped_entry['family_id'],
                mom=maternal_sample,
                dad=paternal_sample,
                sex=Sex.parse(str(ped_entry['sex'])),
                phenotype=ped_entry['affected'] or '0',
            )

    for dataset in cohort.get_datasets():
        samples_with_ped = [s for s in dataset.get_samples() if s.pedigree]
        logging.info(
            f'{dataset.name}: found pedigree info for {len(samples_with_ped)} '
            f'samples out of {len(dataset.get_samples())}'
        )



"""
CPG utils
"""
import pathlib
from typing import Union

import coloredlogs
from cloudpathlib import CloudPath
from cloudpathlib.anypath import to_anypath

# trigger registration for hail-az:// prefixed CloudPaths
import cpg_utils.cloudpath_hail_az

coloredlogs.install(
    level='DEBUG', fmt='%(asctime)s %(levelname)s (%(name)s %(lineno)s): %(message)s'
)

# The AnyPath class https://cloudpathlib.drivendata.org/stable/anypath-polymorphism/
# is very handy to parse a string that can be either a cloud URL or a local posix path.
# However, AnyPath can't be used for type hinting, because neither CloudPath nor
# pathlib.Path derive from it. The AnyPath's constructor method doesn't actually return
# an instance of AnyPath class, but rather Union[CloudPath, pathlib.Path], and it's
# designed to dynamically pick a specific CloudPath or pathlib.Path subclass.
# Here we create an alias for such union to allow using simple "Path" in type hints:
Path = Union[CloudPath, pathlib.Path]

# We would still need to call AnyPath() to parse a string, which might be confusing.
# Something like to_path() would look better, so we are aliasing a handy method
# to_anypath to to_path, which returns exactly the Union type we are looking for:
to_path = to_anypath



"""
Helper functions for working with Git repositories
"""


from typing import List

import os
import re
import subprocess
from shlex import quote


def get_output_of_command(command: List[str], description: str) -> str:
    """
    subprocess.check_output wrapper that returns string output and raises detailed
    exceptions on error.

    Parameters
    ----------
    command:     list of strings creating a full command
    description: meaningful message for error logging

    Returns
    -------
    stripped output string if command was successful

    """
    try:
        return subprocess.check_output(command).decode().strip()
    # Handle and rethrow KeyboardInterrupt error to stop global exception catch
    # pylint: disable=try-except-raise
    except KeyboardInterrupt:
        raise
    except subprocess.CalledProcessError as e:
        raise Exception(
            f"Couldn't call {description} by calling '{' '.join(command)}', {e}"
        ) from e
    except Exception as e:
        raise Exception(
            f"Couldn't process {description} through calling '{' '.join(command)}', {e}"
        ) from e


def get_relative_script_path_from_git_root(script_name: str) -> str:
    """
    If we're in a subdirectory, get the relative path from the git root
    to the current directory, and append the script path.
    For example, the relative path to this script (from git root) is:
        cpg-utils/git.py

    Parameters
    ----------
    script_name

    Returns
    -------
    fully qualified script path from repo root
    """
    base = get_relative_path_from_git_root()
    return os.path.join(base, script_name)


def get_relative_path_from_git_root() -> str:
    """
    If we're in a subdirectory, get the relative path from the git root
    to the current directory. Relpath returns "." if cwd is a git root.
    """
    root = get_git_repo_root()
    return os.path.relpath(os.getcwd(), root)


def get_git_default_remote() -> str:
    """
    Returns the git remote of 'origin',
    e.g. https://github.com/populationgenomics/cpg-utils

    Returns
    -------

    """
    command = ['git', 'remote', 'get-url', 'origin']
    return get_output_of_command(command, 'get Git remote of origin')


def get_git_repo_root() -> str:
    """
    Returns the git repository directory root,
    e.g. /Users/foo/repos/cpg-utils

    Returns
    -------

    """
    command = ['git', 'rev-parse', '--show-toplevel']
    repo_root = get_output_of_command(command, 'get Git repo directory')
    return repo_root


def get_git_commit_ref_of_current_repository() -> str:
    """
    Returns the commit SHA at the current HEAD

    Returns
    -------

    """
    command = ['git', 'rev-parse', 'HEAD']
    return get_output_of_command(command, 'get latest Git commit')


def get_repo_name_from_current_directory() -> str:
    """
    Gets the repo name from the default remote

    Returns
    -------

    """
    return get_repo_name_from_remote(get_git_default_remote())


def get_organisation_name_from_current_directory() -> str:
    """
    Gets the organisation name from the default remote
    Returns
    -------

    """
    return get_organisation_name_from_remote(get_git_default_remote())


def get_organisation_name_from_remote(remote_name: str) -> str:
    """
    Takes the GitHub repo path and obtains the source organisation
    based on its remote URL e.g.:
    >>> get_repo_name_from_remote(\
        'git@github.com:populationgenomics/cpg-utils.git'\
    )
    'populationgenomics'
    >>> get_repo_name_from_remote(\
        'https://github.com/populationgenomics/cpg-utils.git'\
    )
    'populationgenomics'

    Parameters
    ----------
    remote_name

    Returns
    -------
    the organisation name
    """

    organisation = None

    try:
        if remote_name.startswith('http'):
            match = re.match(
                r'http[s]?:\/\/[A-z0-9\.]+?\/(?P<org>.+?)\/.+$', remote_name
            )
            if match:
                organisation = match.group('org')

        elif remote_name.startswith('git@'):
            match = re.match(r'git@[A-z0-9\.]+?:(?P<org>.+?)\/.+$', remote_name)
            if match:
                organisation = match.group('org')
    except AttributeError as ae:
        raise Exception(f'Unsupported remote format: "{remote_name}"') from ae

    if organisation is None:
        raise Exception(f'Unsupported remote format: "{remote_name}"')

    return organisation


def get_repo_name_from_remote(remote_name: str) -> str:
    """
    Get the name of a GitHub repo from a supported organization
    based on its remote URL e.g.:
    >>> get_repo_name_from_remote(\
        'git@github.com:populationgenomics/cpg-utils.git'\
    )
    'cpg-utils'
    >>> get_repo_name_from_remote(\
        'https://github.com/populationgenomics/cpg-utils.git'\
    )
    'cpg-utils'

    removed check for the authorised organisation(s) here - we receive
    a full repository path, so we trust users will not attempt to run
    potentially harmful code
    """

    repo = None
    try:
        if remote_name.startswith('http'):
            match = re.match(
                r'http[s]?:\/\/[A-z0-9\.]+?\/.+?\/(?P<repo>.+)$', remote_name
            )
            if match:
                repo = match.group('repo')

        elif remote_name.startswith('git@'):
            match = re.match(r'git@[A-z0-9\.]+?:.+?\/(?P<repo>.+)$', remote_name)
            if match:
                repo = match.group('repo')
    except AttributeError as ae:
        raise Exception(f'Unsupported remote format: "{remote_name}"') from ae

    if repo is None:
        raise Exception(f'Unsupported remote format: "{remote_name}"')

    if repo.endswith('.git'):
        repo = repo[:-4]

    return repo


def check_if_commit_is_on_remote(commit: str) -> bool:
    """
    Returns 'True' if the commit is available on a remote branch.
    This relies on the current environment to be up-to-date.
    It asks if the local environment knows a remote branch with the commit.

    Parameters
    ----------
    commit

    Returns
    -------

    """
    command = ['git', 'branch', '-r', '--contains', commit]
    try:
        ret = subprocess.check_output(command)
        return bool(ret)
    except subprocess.CalledProcessError:
        return False


def prepare_git_job(
    job,
    organisation: str,
    repo_name: str,
    commit: str,
    is_test: bool = True,
    print_all_statements: bool = True,
):
    """
    Takes a hail batch job, and:
        * Clones the repository
            * if access_level != "test": check the desired commit is on 'main'
        * Check out the specific commit

    Parameters
    ----------
    job                     - A hail BashJob
    organisation            - The GitHub individual or organisation
    repo_name               - The repository name to check out
    commit                  - The commit hash to check out
    is_test                 - CPG specific: only Main commits can run on Main data
    print_all_statements    - logging toggle

    Returns
    -------
    No return required
    """

    # Use "set -x" to print the commands for easier debugging.
    if print_all_statements:
        job.command('set -x')

    # Note: for private GitHub repos we'd need to use a token to clone.
    # Any job commands here are evaluated in a bash shell, so user arguments should
    # be escaped to avoid command injection.
    job.command(
        f'git clone --recurse-submodules https://github.com/{quote(organisation)}/{quote(repo_name)}.git'
    )
    job.command(f'cd {quote(repo_name)}')

    # Except for the "test" access level, we check whether commits have been
    # reviewed by verifying that the given commit is in the main branch.
    if not is_test:
        job.command('git checkout main')
        job.command(
            f'git merge-base --is-ancestor {quote(commit)} HEAD || '
            '{ echo "error: commit not merged into main branch"; exit 1; }'
        )
    job.command(f'git checkout {quote(commit)}')
    job.command(f'git submodule update')



from abc import ABC, abstractmethod
from typing import Dict, Optional

import azure.core.exceptions
import azure.identity
import azure.storage.blob
import google.cloud.storage

from .creds import get_azure_credentials
from .deploy_config import get_deploy_config, get_server_config

data_manager: "DataManager" = None

class DataManager(ABC):
    """Multi-cloud abstraction for reading/writing cloud dataset blobs."""

    @staticmethod
    def get_data_manager(cloud_type: Optional[str] = None) -> "DataManager":
        """Instantiates a DataManager object of the appropriate cloud type."""
        if not cloud_type:
            cloud_type = get_deploy_config().cloud
        if cloud_type == "azure":
            return DataManagerAzure()
        assert cloud_type == "gcp"
        return DataManagerGCP()

    @abstractmethod
    def get_dataset_bucket_url(self, dataset: str, bucket_type: str) -> str:
        """Build dataset-specific bucket URL with Hail-style scheme ("gs:" or "hail-az:")."""

    @abstractmethod
    def get_blob(self, dataset: Optional[str], bucket_type: str, blob_path: str) -> Optional[bytes]:
        """Cloud-specific blob read. Returns None if blob doesn't exist."""

    @abstractmethod
    def set_blob(self, dataset: str, bucket_type: str, blob_path: str, contents: bytes) -> None:
        """Cloud-specific blob write."""


class DataManagerGCP(DataManager):
    """GCP Storage wrapper for reading/writing cloud dataset blobs."""

    _storage_client: google.cloud.storage.Client = None

    def __init__(self):
        """Loads GCP credentials and caches storage client."""
        self._storage_client = google.cloud.storage.Client()

    def get_dataset_bucket_url(self, dataset: str, bucket_type: str) -> str:
        """Build dataset-specific Hail-style bucket URL for GCP ("gs://...")."""
        return f"gs://cpg-{dataset}-{bucket_type}"

    def get_blob(self, dataset: Optional[str], bucket_type: str, blob_path: str) -> Optional[bytes]:
        """Reads a GCP storage bucket blob."""
        bucket_name = f"cpg-{dataset}-{bucket_type}" if dataset else f"cpg-{bucket_type}"
        bucket = self._storage_client.bucket(bucket_name)
        blob = bucket.get_blob(blob_path)
    
        return None if blob is None else blob.download_as_bytes()

    def set_blob(self, dataset: Optional[str], bucket_type: str, blob_path: str, contents: bytes) -> None:
        """Writes a GCP storage bucket blob."""
        bucket_name = f"cpg-{dataset}-{bucket_type}" if dataset else f"cpg-{bucket_type}"
        bucket = self._storage_client.bucket(bucket_name)
        blob = bucket.blob(blob_path)
        with blob.open(mode="wb") as f:
            f.write(contents)


class DataManagerAzure(DataManager):
    """Azure Storage wrapper for reading/writing cloud dataset blobs."""

    _credential = None

    def __init__(self):
        self._credential = get_azure_credentials()

    def get_storage_account(self, dataset: Optional[str]) -> str:
        """Gets storage host account name based on dataset name or AR base (without scheme)."""
        if dataset:
            # Need to map dataset name to storage account name.
            server_config = get_server_config()
            if dataset not in server_config:
                raise ValueError(f"No such dataset in server config: {dataset}")
            account = server_config[dataset]["projectId"]
        else: # Otherwise use the base analysis_runner storage account.
            account = get_deploy_config().analysis_runner_project
        return f"{account}sa"

    def get_dataset_bucket_url(self, dataset: str, bucket_type: str) -> str:
        """Build dataset-specific Hail-style bucket URL for Azure ("hail-az://...")."""
        return f"hail-az://{self.get_storage_account(dataset)}/cpg-{dataset}-{bucket_type}"

    def get_blob(self, dataset: Optional[str], bucket_type: str, blob_path: str) -> Optional[bytes]:
        """Reads an Azure storage blob."""
        storage_url = "https://" + self.get_storage_account(dataset) + ".blob.core.windows.net"
        container_name = f"cpg-{dataset}-{bucket_type}" if dataset else f"cpg-{bucket_type}"
        blob_client = azure.storage.blob.BlobClient(
            storage_url, container_name, blob_path, credential=self._credential
        )
        try:
            download_stream = blob_client.download_blob()
        except azure.core.exceptions.ResourceNotFoundError:
            return None

        return download_stream.readall()

    def set_blob(self, dataset: Optional[str], bucket_type: str, blob_path: str, contents: bytes) -> None:
        """Writes an Azure storage blob."""
        storage_url = "https://" + self.get_storage_account(dataset) + ".blob.core.windows.net"
        container_name = f"cpg-{dataset}-{bucket_type}" if dataset else f"cpg-{bucket_type}"
        blob_client = azure.storage.blob.BlobClient(
            storage_url, container_name, blob_path, credential=self._credential
        )
        blob_client.upload_blob(data=contents, overwrite=True)


def get_data_manager() -> DataManager:
    global data_manager
    if data_manager is None:
        data_manager = DataManager.get_data_manager()
    return data_manager


def clear_data_manager() -> None:
    global data_manager
    data_manager = None


def get_dataset_bucket_url(dataset: str, bucket_type: str) -> str:
    """Return dataset-specific bucket URL with Hail-style scheme ("gs:" or "hail-az:")."""
    return get_data_manager().get_dataset_bucket_url(dataset, bucket_type)



import os
from unittest.mock import patch

import pytest
from azure.identity import ClientSecretCredential, DefaultAzureCredential
from cpg_utils.cloudpath_hail_az import HailAzureBlobPath, HailAzureBlobClient
import cloudpathlib.exceptions

def test_account_url():
    with pytest.raises(ValueError):
        HailAzureBlobClient(account_url=None)


@patch('cpg_utils.cloudpath_hail_az.BlobServiceClient', autospec=True)
def test_sas_auth(mock_blob_service_client):
    path = HailAzureBlobPath('hail-az://fakeaccount/fakecontainer/fakeblob?not_a_real_sas_token')
    assert mock_blob_service_client.call_args.kwargs['account_url'].endswith('?not_a_real_sas_token')


@patch('cpg_utils.cloudpath_hail_az.BlobServiceClient', autospec=True)
def test_env_auth(mock_blob_service_client, test_resources_path):
    with patch.dict(os.environ, {"AZURE_APPLICATION_CREDENTIALS": os.path.join(test_resources_path, 'azure_creds.json')}):
        path = HailAzureBlobPath('hail-az://fakeaccount/fakecontainer/fakeblob')
        assert isinstance(mock_blob_service_client.call_args.kwargs['credential'], ClientSecretCredential)


@patch('cpg_utils.cloudpath_hail_az.BlobServiceClient', autospec=True)
def test_default_auth(mock_blob_service_client):
    path = HailAzureBlobPath('hail-az://fakeaccount/fakecontainer/fakeblob')
    assert isinstance(mock_blob_service_client.call_args.kwargs['credential'], DefaultAzureCredential)


def test_valid_cloudpath():
    with pytest.raises(cloudpathlib.exceptions.InvalidPrefixError):
        HailAzureBlobPath.is_valid_cloudpath("foo://bar/baz/bah", raise_on_error=True)
    with pytest.raises(cloudpathlib.exceptions.InvalidPrefixError):
        HailAzureBlobPath.is_valid_cloudpath("hail-az://bar.blob.core.windows.net/baz/bah", raise_on_error=True)
    with pytest.raises(cloudpathlib.exceptions.InvalidPrefixError):
        HailAzureBlobPath.is_valid_cloudpath("https://bar/baz/bah", raise_on_error=True)
    with pytest.raises(cloudpathlib.exceptions.InvalidPrefixError):
        HailAzureBlobPath.is_valid_cloudpath("https://bar.blob.core.windows.COM/baz/bah", raise_on_error=True)


@patch('cpg_utils.cloudpath_hail_az.BlobServiceClient', autospec=True)
def test_client_provided( _ ):
    client = HailAzureBlobClient('https://fakeaccount.blob.core.windows.net')
    path = HailAzureBlobPath('hail-az://fakeaccount/fakecontainer/fakeblob', client=client)
    assert path.client == client


@patch('cpg_utils.cloudpath_hail_az.BlobServiceClient', autospec=True)
def test_cloudpath_provided( _ ):
    path1 = HailAzureBlobPath('hail-az://fakeaccount/fakecontainer')
    path2 = HailAzureBlobPath(path1)
    assert path1.as_uri() == path2.as_uri()


@patch('cpg_utils.cloudpath_hail_az.BlobServiceClient', autospec=True)
def test_path_completeness(monkeypatch):
    with pytest.raises(ValueError):
        path = HailAzureBlobPath('hail-az://fakeaccount')
    with pytest.raises(ValueError):
        path = HailAzureBlobPath('hail-az://fakeaccount/')
    HailAzureBlobPath('hail-az://fakeaccount/fakecontainer')
    HailAzureBlobPath('hail-az://fakeaccount/fakecontainer/fakeblob')
    HailAzureBlobPath('hail-az://fakeaccount/fakecontainer/fakeblob/')
    HailAzureBlobPath('hail-az://fakeaccount/fakecontainer/fakeblob/subpath')
    HailAzureBlobPath('hail-az://fakeaccount/fakecontainer/fakeblob/subpath/')

@patch('cpg_utils.cloudpath_hail_az.BlobServiceClient', autospec=True)
def test_properties(monkeypatch):
    path = HailAzureBlobPath('hail-az://fakeaccount/fakecontainer')
    assert path.account == 'fakeaccount'
    assert path.container == 'fakecontainer'
    assert not path.blob
    path = HailAzureBlobPath('hail-az://fakeaccount/fakecontainer/')
    assert not path.blob
    path = HailAzureBlobPath('hail-az://fakeaccount/fakecontainer/fakeblob')
    assert path.blob == 'fakeblob' 
    


"""Convenience functions related to cloud infrastructure."""

from typing import Optional
import traceback
import logging
import google.auth
from google.auth.exceptions import DefaultCredentialsError
from google.auth.transport.requests import Request
from google.cloud import secretmanager
from google.oauth2 import id_token
from google.auth import jwt
import google.api_core.exceptions


def email_from_id_token(id_token_jwt: str) -> str:
    """Decodes the ID token (JWT) to get the email address of the caller.

    See http://bit.ly/2YAIkzy for details.

    This function assumes that the token has been verified beforehand."""

    return jwt.decode(id_token_jwt, verify=False)['email']


def read_secret(
    project_id: str,
    secret_name: str,
    fail_gracefully: bool = True,
) -> Optional[str]:
    """Reads the latest version of a GCP Secret Manager secret.

    Returns None if the secret doesn't exist or there was a problem retrieving it,
    unless `fail_gracefully` is set to False."""

    secret_manager = secretmanager.SecretManagerServiceClient()
    secret_path = secret_manager.secret_version_path(project_id, secret_name, 'latest')

    try:
        # noinspection PyTypeChecker
        response = secret_manager.access_secret_version(request={'name': secret_path})
        return response.payload.data.decode('UTF-8')
    except google.api_core.exceptions.ClientError:
        # Fail gracefully if there's no secret version yet.
        if fail_gracefully:
            traceback.print_exc()
            return None
        raise
    except AttributeError:
        # Sometimes the google API fails when no version is present, with:
        #   File "{site-packages}/google/api_core/exceptions.py",
        #   line 532, in from_grpc_error if isinstance(rpc_exc, grpc.Call) or _is_informative_grpc_error(rpc_exc):
        #   AttributeError: 'NoneType' object has no attribute 'Call'
        if fail_gracefully:
            traceback.print_exc()
            return None
        raise


def write_secret(project_id: str, secret_name: str, secret_value: str) -> None:
    """
    Adds a new version for a GCP Secret Manager secret and disables all previous versions

    Parameters
    ----------
    project_id
    secret_name
    secret_value

    Returns
    -------

    """

    secret_manager = secretmanager.SecretManagerServiceClient()
    secret_path = secret_manager.secret_path(project_id, secret_name)

    response = secret_manager.add_secret_version(
        request={
            'parent': secret_path,
            'payload': {'data': secret_value.encode('UTF-8')},
        }
    )

    # Disable all previous versions.
    for version in secret_manager.list_secret_versions(request={'parent': secret_path}):
        # Don't attempt to change the state of destroyed / already disabled secrets and
        # don't disable the latest version.
        if (
            version.state == secretmanager.SecretVersion.State.ENABLED
            and version.name != response.name
        ):
            secret_manager.disable_secret_version(request={'name': version.name})


def get_google_identity_token(audience: str) -> str:
    """
    Returns a Google identity token for the given audience.
    """

    # Suppress DEBUG / WARNING messages from google-auth library,
    # in case we're not running on GCE. This gets restored at the end.
    logger = logging.getLogger()
    previous_level = logger.getEffectiveLevel()
    logger.setLevel(logging.ERROR)

    # Need to support two distinct cases:
    # - service accounts on GCE VMs (default),
    # - human users with default credentials.
    try:
        return id_token.fetch_id_token(Request(), audience=audience)
    except DefaultCredentialsError:
        # https://stackoverflow.com/a/55804230
        creds, _ = google.auth.default()
        creds.refresh(Request())
        return creds.id_token
    finally:
        logger.setLevel(previous_level)



import json
import logging
from os import getenv
from typing import Any, Dict, Optional
from .secrets import SecretManager

deploy_config: "DeployConfig" = None
DEFAULT_CONFIG = {
    "cloud": "gcp",
    "sample_metadata_project": "sample-metadata",
    "sample_metadata_host": "http://localhost:8000",
    "analysis_runner_project": "analysis-runner",
    "analysis_runner_host": "http://localhost:8001",
    "container_registry": "australia-southeast1-docker.pkg.dev",
    "web_host_base": "web.populationgenomics.org.au"
}


class DeployConfig:

    _server_config: Dict[str, Any] = None
    _secret_manager: SecretManager = None

    @staticmethod
    def from_dict(config: Dict[str, str]) -> "DeployConfig":
        return DeployConfig(**config)

    @staticmethod
    def from_environment() -> "DeployConfig":
        deploy_config = json.loads(getenv("CPG_DEPLOY_CONFIG", json.dumps(DEFAULT_CONFIG)))
        # Allow individual field overrides.
        deploy_config["cloud"] = getenv("CLOUD", deploy_config["cloud"])
        deploy_config["sample_metadata_host"] = getenv("SM_HOST_URL", deploy_config["sample_metadata_host"])
        return DeployConfig.from_dict(deploy_config)

    def __init__(
        self,
        cloud: Optional[str],
        sample_metadata_project: Optional[str],
        sample_metadata_host: Optional[str],
        analysis_runner_project: Optional[str],
        analysis_runner_host: Optional[str],
        container_registry: Optional[str],
        web_host_base: Optional[str],
    ):
        self.cloud = cloud or DEFAULT_CONFIG["cloud"]
        self.sample_metadata_project = sample_metadata_project or DEFAULT_CONFIG["sample_metadata_project"]
        self.sample_metadata_host = sample_metadata_host or DEFAULT_CONFIG["sample_metadata_host"]
        self.analysis_runner_project = analysis_runner_project or DEFAULT_CONFIG["analysis_runner_project"]
        self.analysis_runner_host = analysis_runner_host or DEFAULT_CONFIG["analysis_runner_host"]
        self.container_registry = container_registry or DEFAULT_CONFIG["container_registry"]
        self.web_host_base = web_host_base or DEFAULT_CONFIG["web_host_base"]
        assert self.cloud in ("gcp", "azure"), f"Invalid cloud specification '{self.cloud}'"

    def to_dict(self) -> Dict[str, str]:
        return self.__dict__.copy()

    @property
    def secret_manager(self) -> SecretManager:
        if self._secret_manager is None:
            self._secret_manager = SecretManager.get_secret_manager(self.cloud)
        return self._secret_manager

    @property
    def server_config(self) -> Dict[str, Any]:
        if self._server_config is None:
            config = self.read_global_config("server-config")
            self._server_config = json.loads(config)
        return self._server_config

    def read_project_id_config(self, project_id: str, config_key: str) -> str:
        config_host = project_id + "vault" if self.cloud == "azure" else project_id
        return self.secret_manager.read_secret(config_host, config_key)

    def read_global_config(self, config_key: str) -> str:
        return self.read_project_id_config(self.analysis_runner_project, config_key)

    def read_dataset_config(self, dataset: str, config_key: str) -> str:
        if dataset not in self.server_config:
            return ""
        dataset_id = self.server_config[dataset]["projectId"]
        return self.read_project_id_config(dataset_id, config_key)


def get_deploy_config() -> DeployConfig:
    global deploy_config
    if deploy_config is None:
        set_deploy_config_from_env()
    return deploy_config


def set_deploy_config(config: DeployConfig) -> None:
    global deploy_config
    logging.info(f"setting deploy_config: {json.dumps(config.to_dict())}")
    deploy_config = config


def set_deploy_config_from_env() -> None:
    set_deploy_config(DeployConfig.from_environment())


def get_server_config() -> Dict[str, Any]:
    return get_deploy_config().server_config



import os

import azure.identity
import google.auth
import google.auth.exceptions
import google.auth.transport.requests
import pytest
from cpg_utils.auth import get_user_from_headers
from cpg_utils.creds import get_sample_metadata_token
from cpg_utils.deploy_config import set_deploy_config_from_env
from google.oauth2 import service_account

# Mocked tokens from https://www.javainuse.com/jwtgenerator
TEST_TOKEN1 = "eyJhbGciOiJIUzI1NiJ9.eyJSb2xlIjoiQWRtaW4iLCJJc3N1ZXIiOiJJc3N1ZXIiLCJVc2VybmFtZSI6IkphdmFJblVzZSIsImV4cCI6MTY1MDczNDA0NSwiaWF0IjoxNjUwNzM0MDQ1LCJlbWFpbCI6InRlc3QxQHRlc3QuY29tIn0.OJ-39xdDbIH8FDsdlwFsIwyDzgSbA_gOtYbRNhBmLxo"
TEST_TOKEN2 = "eyJhbGciOiJIUzI1NiJ9.eyJSb2xlIjoiQWRtaW4iLCJJc3N1ZXIiOiJJc3N1ZXIiLCJVc2VybmFtZSI6IkphdmFJblVzZSIsImV4cCI6MTY1MDczNDA0NSwiaWF0IjoxNjUwNzM0MDQ1LCJlbWFpbCI6InRlc3QyQHRlc3QuY29tIn0.ONbKZ4cf0jb9wtVJBprdtRhAhc5KVp9hSAPWN6ukt9A"
TEST_TOKEN3 = "eyJhbGciOiJIUzI1NiJ9.eyJSb2xlIjoiQWRtaW4iLCJJc3N1ZXIiOiJJc3N1ZXIiLCJVc2VybmFtZSI6IkphdmFJblVzZSIsImV4cCI6MTY1MDczNDA0NSwiaWF0IjoxNjUwNzM0MDQ1LCJlbWFpbCI6InRlc3QzQHRlc3QuY29tIn0.njCokays7b_Yl2O0_1lKROvLV-MiA0RW4bwx68dqeTo"


def test_bogus_header(monkeypatch):
    assert get_user_from_headers({}) is None
    headers = { "x-goog-iap-jwt-assertion" : "bogus" }
    monkeypatch.setenv("CLOUD", "gcp")
    set_deploy_config_from_env()
    with pytest.raises(ValueError, match="Wrong number of segments in token"):
        get_user_from_headers(headers)
    headers = { "Authorization" : "Bearer bogus" }
    with pytest.raises(ValueError, match="Wrong number of segments in token"):
        get_user_from_headers(headers)


def test_headers(monkeypatch):
    headers = {
        "x-goog-iap-jwt-assertion" : TEST_TOKEN1,
        "x-ms-client-principal-name" : "test2@test.com",
        "Authorization" : "Bearer " + TEST_TOKEN3
    }
    monkeypatch.setenv("CLOUD", "gcp")
    set_deploy_config_from_env()
    assert get_user_from_headers(headers) == "test1@test.com"
    monkeypatch.setenv("CLOUD", "azure")
    set_deploy_config_from_env()
    assert get_user_from_headers(headers) == "test2@test.com"
    headers = {
        "x-goog-iap-jwt-assertion" : TEST_TOKEN1,
        "Authorization" : "Bearer " + TEST_TOKEN3
    }
    assert get_user_from_headers(headers) == "test3@test.com"


class MockCredentials():
    def __init__(self, **kwargs):
        assert kwargs.get("tenant", "TENANT") == "TENANT"
        assert kwargs.get("appId", "APPID") == "APPID"
        assert kwargs.get("password", "supersecret") == "supersecret"
        self.token = "creds"
        self.id_token = "gcpcredsdef"

    def get_token(self, scope):
        assert scope == "api://smapi-sample-metadata/.default"
        return self

    def refresh(self, auth_req):
        pass


def mock_get_creds(*args, **kwargs):
    assert kwargs.get("target_audience", "http://localhost:8000") == "http://localhost:8000"
    return MockCredentials()


def mock_get_default_creds():
    return MockCredentials(), 42


def test_az_default_token(monkeypatch):
    monkeypatch.setenv("CLOUD", "azure")
    monkeypatch.delenv("CPG_DEPLOY_CONFIG", raising=False)
    monkeypatch.setattr(azure.identity, "DefaultAzureCredential", mock_get_creds)
    set_deploy_config_from_env()
    assert get_sample_metadata_token() == "creds"


def test_az_file_token(monkeypatch, test_resources_path):
    monkeypatch.delenv("CPG_DEPLOY_CONFIG", raising=False)
    monkeypatch.setenv("CLOUD", "azure")
    set_deploy_config_from_env()
    monkeypatch.setattr(azure.identity, "ClientSecretCredential", mock_get_creds)
    monkeypatch.setenv("AZURE_APPLICATION_CREDENTIALS", os.path.join(test_resources_path, "azure_creds.json"))
    assert get_sample_metadata_token() == "creds"


def test_gcp_default_token(monkeypatch):
    monkeypatch.setenv("CLOUD", "gcp")
    set_deploy_config_from_env()
    monkeypatch.setattr(google.auth, "default", mock_get_default_creds)
    assert get_sample_metadata_token() == "gcpcredsdef"


def test_gcp_file_token(monkeypatch, test_resources_path):
    monkeypatch.delenv("CPG_DEPLOY_CONFIG", raising=False)
    monkeypatch.setenv("CLOUD", "gcp")
    set_deploy_config_from_env()
    monkeypatch.setattr(service_account.IDTokenCredentials, "from_service_account_info", mock_get_creds)
    monkeypatch.setenv("GOOGLE_APPLICATION_CREDENTIALS", os.path.join(test_resources_path, "gcp_creds.json"))
    assert get_sample_metadata_token() == "creds"



"""
Wrappers for bioinformatics file types (CRAM, GVCF, FASTQ, etc).
"""

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Union

from hailtop.batch import ResourceGroup, ResourceFile, Batch

from cpg_utils import Path, to_path
from cpg_utils.workflows.utils import exists


class AlignmentInput(ABC):
    """
    Data that works as input for alignment or realignment.
    """

    @abstractmethod
    def exists(self) -> bool:
        """
        Check if all files exist.
        """


class CramOrBamPath(AlignmentInput, ABC):
    """
    Represents a path to a CRAM or a BAM file, optionally with corresponding index.
    """

    def __init__(
        self,
        path: str | Path,
        index_path: str | Path | None = None,
        reference_assembly: str = None,
    ):
        self.path = to_path(path)
        self.index_path: Path | None = None
        self.full_index_suffix: str | None = None
        if index_path:
            self.index_path = to_path(index_path)
            assert self.index_path.suffix == f'.{self.index_ext}'
            self.full_index_suffix = str(self.index_path).replace(
                str(self.path.with_suffix('')), ''
            )

    @property
    @abstractmethod
    def ext(self) -> str:
        ...

    @property
    @abstractmethod
    def index_ext(self) -> str:
        ...

    def __str__(self) -> str:
        return str(self.path)

    def __repr__(self) -> str:
        """
        >>> repr(CramPath('gs://bucket/sample.cram', 'gs://bucket/sample.cram.crai'))
        'CRAM(gs://bucket/sample.cram+.cram.crai)'
        """
        res = str(self.path)
        if self.index_path:
            assert self.full_index_suffix
            res += f'+{self.full_index_suffix}'
        return f'{self.ext.upper()}({res})'

    def exists(self) -> bool:
        """
        CRAM file exists.
        """
        return exists(self.path)

    def resource_group(self, b: Batch) -> ResourceGroup:
        """
        Create a Hail Batch resource group
        """
        d = {
            self.ext: str(self.path),
        }
        if self.full_index_suffix:
            d[self.full_index_suffix] = str(self.index_path)

        return b.read_input_group(**d)


class BamPath(CramOrBamPath):
    """
    Represents a path to a BAM file, optionally with corresponding index.
    """

    EXT = 'bam'
    INDEX_EXT = 'bai'

    def __init__(
        self,
        path: str | Path,
        index_path: str | Path | None = None,
    ):
        super().__init__(path, index_path)

    @property
    def ext(self) -> str:
        return BamPath.EXT

    @property
    def index_ext(self) -> str:
        return BamPath.INDEX_EXT


class CramPath(CramOrBamPath):
    """
    Represents a path to a CRAM file, optionally with corresponding index.
    """

    EXT = 'cram'
    INDEX_EXT = 'crai'

    def __init__(
        self,
        path: str | Path,
        index_path: str | Path | None = None,
        reference_assembly: str | Path = None,
    ):
        self.reference_assembly = None
        if reference_assembly:
            self.reference_assembly = to_path(reference_assembly)
        super().__init__(path, index_path)
        self.somalier_path = to_path(f'{self.path}.somalier')

    @property
    def ext(self) -> str:
        return CramPath.EXT

    @property
    def index_ext(self) -> str:
        return CramPath.INDEX_EXT


class GvcfPath:
    """
    Represents GVCF data on a bucket within the workflow.
    Includes a path to a GVCF file along with a corresponding TBI index,
    and a corresponding fingerprint path.
    """

    def __init__(self, path: Path | str):
        self.path = to_path(path)
        self.somalier_path = to_path(f'{self.path}.somalier')

    def __str__(self) -> str:
        return str(self.path)

    def __repr__(self) -> str:
        return f'GVCF({self.path})'

    def exists(self) -> bool:
        """
        GVCF file exists.
        """
        return self.path.exists()

    @property
    def tbi_path(self) -> Path:
        """
        Path to the corresponding index
        """
        return to_path(f'{self.path}.tbi')

    def resource_group(self, b: Batch) -> ResourceGroup:
        """
        Create a Hail Batch resource group
        """
        return b.read_input_group(
            **{
                'g.vcf.gz': str(self.path),
                'g.vcf.gz.tbi': str(self.tbi_path),
            }
        )


FastqPath = Union[str, Path, ResourceFile]


@dataclass
class FastqPair(AlignmentInput):
    """
    Pair of FASTQ files
    """

    r1: FastqPath
    r2: FastqPath

    def __getitem__(self, i):
        assert i == 0 or i == 1, i
        return [self.r1, self.r2][i]

    def as_resources(self, b) -> 'FastqPair':
        """
        Makes a pair of ResourceFile objects for r1 and r2.
        """
        return FastqPair(
            *[
                self[i]
                if isinstance(self[i], ResourceFile)
                else b.read_input(str(self[i]))
                for i in [0, 1]
            ]
        )

    def exists(self) -> bool:
        """
        Check if each FASTQ file in the pair exists.
        """
        return exists(self.r1) and exists(self.r2)

    def __repr__(self):
        """
        Glob string to find all FASTQ files.

        >>> str(FastqPair('gs://sample_R1.fq.gz', 'gs://sample_R2.fq.gz'))
        'gs://sample_R{1,2}.fq.gz'
        """
        return ''.join(
            f'{{{",".join(sorted(set(chars)))}}}' if len(set(chars)) > 1 else chars[0]
            for chars in zip(str(self.r1), str(self.r2))
        )


class FastqPairs(list[FastqPair], AlignmentInput):
    """
    Multiple FASTQ file pairs belonging to the same sample
    (e.g. multiple lanes or top-ups).
    """

    def exists(self) -> bool:
        """
        Check if each FASTQ file in each pair exists.
        """
        return all(pair.exists() for pair in self)

    def __repr__(self) -> str:
        """
        Glob string to find all FASTQ files.

        >>> repr(FastqPairs([FastqPair('gs://sample_R1.fq.gz', 'gs://sample_R2.fq.gz')]))
        'gs://sample_R{1,2}.fq.gz'
        >>> p1 = FastqPair('gs://sample_L2_R1.fq.gz', 'gs://sample_L2_R2.fq.gz')
        >>> p2 = FastqPair('gs://sample_L1_R1.fq.gz', 'gs://sample_L1_R2.fq.gz')
        >>> repr(FastqPairs([p1, p2]))
        'gs://sample_L{1,2}_R{1,2}.fq.gz'
        """
        return ''.join(
            f'{{{",".join(sorted(set(chars)))}}}' if len(set(chars)) > 1 else chars[0]
            for chars in zip(*[repr(pair) for pair in self])
        )



import os
import json

import pytest


@pytest.fixture
def test_resources_path():
    return os.path.abspath(os.path.join(os.path.dirname(__file__), "resources"))

@pytest.fixture
def json_load(test_resources_path):
    def _loader(file_name):
        with open(os.path.join(test_resources_path, file_name), "r") as json_file:
            return json.load(json_file)
    return _loader



#!/usr/bin/env python

from setuptools import find_packages, setup

setup(
    name='cpg-utils-ms',
    # This tag is automatically updated by bumpversion
    version='0.11.1',
    description='Library of convenience functions specific to the CPG (MS version)',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url=f'https://github.com/gregsmi/cpg-utils',
    license='MIT',
    packages=find_packages(),
    install_requires=[
        'google-auth',
        'google-cloud-secret-manager',
        'azure-identity',
        'azure-keyvault-secrets',
        'azure-storage-blob',
        'cloudpathlib[all]',
        'toml',
        'frozendict',
        'coloredlogs',
    ],
    extras_require={
        'test': [
            'pytest',
            'pytest-mock',
            'pytest-cov'
        ],
        'workflows': [
            'hail',
            'networkx',
            'sample-metadata-ms',
        ],
    },
    package_data={
        'cpg_utils': ['config-template.toml'],
    },
    keywords='bioinformatics',
    classifiers=[
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX',
        'Operating System :: Unix',
        'Programming Language :: Python',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
    ],
)
